"""Smooth background quantities

CCL defines seven species types:

- 'matter': cold dark matter and baryons
- 'dark_energy': cosmological constant or otherwise
- 'radiation': relativistic species besides massless neutrinos (i.e.,
  only photons)
- 'curvature': curvature density
- 'neutrinos_rel': relativistic neutrinos
- 'neutrinos_massive': massive neutrinos

These strings define the `species` inputs to the functions below.
"""
import numpy as np
from . import ccllib as lib
from .pyutils import _vectorize_fn, _vectorize_fn3
from .pyutils import _vectorize_fn4, _vectorize_fn5
from .parameters import physical_constants

species_types = {
    'critical': lib.species_crit_label,
    'matter': lib.species_m_label,
    'dark_energy': lib.species_l_label,
    'radiation': lib.species_g_label,
    'curvature': lib.species_k_label,
    'neutrinos_rel': lib.species_ur_label,
    'neutrinos_massive': lib.species_nu_label,
}


def growth_factor(cosmo, a):
    """Growth factor.

    Args:
        cosmo (:class:`~pyccl.core.Cosmology`): Cosmological parameters.
        a (float or array_like): Scale factor(s), normalized to 1 today.

    Returns:
        float or array_like: Growth factor, normalized to unity today.
    """
    cosmo.compute_growth()
    return _vectorize_fn(lib.growth_factor,
                         lib.growth_factor_vec, cosmo, a)


def growth_factor_unnorm(cosmo, a):
    """Unnormalized growth factor.

    Args:
        cosmo (:class:`~pyccl.core.Cosmology`): Cosmological parameters.
        a (float or array_like): Scale factor(s), normalized to 1 today.

    Returns:
        float or array_like: Unnormalized growth factor, normalized to \
            the scale factor at early times.
    """
    cosmo.compute_growth()
    return _vectorize_fn(lib.growth_factor_unnorm,
                         lib.growth_factor_unnorm_vec, cosmo, a)


def growth_rate(cosmo, a):
    """Growth rate defined as the logarithmic derivative of the
    growth factor, dlnD/dlna.

    Args:
        cosmo (:class:`~pyccl.core.Cosmology`): Cosmological parameters.
        a (float or array_like): Scale factor(s), normalized to 1 today.

    Returns:
        float or array_like: Growth rate.

    """
    cosmo.compute_growth()
    return _vectorize_fn(lib.growth_rate,
                         lib.growth_rate_vec, cosmo, a)


def comoving_radial_distance(cosmo, a):
    """Comoving radial distance.

    Args:
        cosmo (:class:`~pyccl.core.Cosmology`): Cosmological parameters.
        a (float or array_like): Scale factor(s), normalized to 1 today.

    Returns:
        float or array_like: Comoving radial distance; Mpc.
    """
    cosmo.compute_distances()
    return _vectorize_fn(lib.comoving_radial_distance,
                         lib.comoving_radial_distance_vec, cosmo, a)


def comoving_angular_distance(cosmo, a):
    """Comoving angular distance.

    .. note:: This quantity is otherwise known as the transverse
              comoving distance, and is NOT angular diameter
              distance or angular separation. The comoving angular distance
              is defined such that the comoving distance between
              two objects at a fixed scale factor separated by an angle
              :math:`\\theta` is :math:`\\theta r_{A}(a)` where
              :math:`r_{A}(a)` is the comoving angular distance.

    Args:
        cosmo (:class:`~pyccl.core.Cosmology`): Cosmological parameters.
        a (float or array_like): Scale factor(s), normalized to 1 today.

    Returns:
        float or array_like: Comoving angular distance; Mpc.
    """
    cosmo.compute_distances()
    return _vectorize_fn(lib.comoving_angular_distance,
                         lib.comoving_angular_distance_vec, cosmo, a)


def angular_diameter_distance(cosmo, a1, a2=None):
    """Angular diameter distance.

    The angular diameter distance in Mpc from scale factor
    `a1` to scale factor `a2`. If `a2` is not provided, it is
    assumed that the distance will be calculated between 1 and
    `a1`.

    .. note:: `a2` has to be smaller than `a1` (i.e. a source at
              `a2` is behind one at `a1`). You can compute the
              distance between a single lens at `a1` and multiple
              sources at `a2` by passing a scalar `a1`.

    Args:
        cosmo (:class:`~pyccl.core.Cosmology`): Cosmological parameters.
        a1 (float or array_like): Scale factor(s), normalized to 1 today.
        a2 (float or array_like): Scale factor(s), normalized to 1 today,
        optional.

    Returns:
        float or array_like: angular diameter distance; Mpc.
    """
    cosmo.compute_distances()
    if (a2 is not None):
        # One lens, multiple sources
        if (np.ndim(a1) == 0) and (np.ndim(a2) != 0):
            a1 = np.full(len(a2), a1)
        return _vectorize_fn5(lib.angular_diameter_distance,
                              lib.angular_diameter_distance_vec,
                              cosmo, a1, a2)
    else:
        if isinstance(a1, (int, float)):
            return _vectorize_fn5(lib.angular_diameter_distance,
                                  lib.angular_diameter_distance_vec,
                                  cosmo, 1., a1)
        else:
            return _vectorize_fn5(lib.angular_diameter_distance,
                                  lib.angular_diameter_distance_vec,
                                  cosmo, np.ones(len(a1)), a1)


def h_over_h0(cosmo, a):
    """Ratio of Hubble constant at `a` over Hubble constant today.

    Args:
        cosmo (:class:`~pyccl.core.Cosmology`): Cosmological parameters.
        a (float or array_like): Scale factor(s), normalized to 1 today.

    Returns:
        float or array_like: H(a)/H0.
    """
    cosmo.compute_distances()
    return _vectorize_fn(lib.h_over_h0,
                         lib.h_over_h0_vec, cosmo, a)


def luminosity_distance(cosmo, a):
    """Luminosity distance.

    Args:
        cosmo (:class:`~pyccl.core.Cosmology`): Cosmological parameters.
        a (float or array_like): Scale factor(s), normalized to 1 today.

    Returns:
        float or array_like: Luminosity distance; Mpc.
    """
    cosmo.compute_distances()
    return _vectorize_fn(lib.luminosity_distance,
                         lib.luminosity_distance_vec, cosmo, a)


def distance_modulus(cosmo, a):
    """Distance Modulus, defined as 5 * log10(luminosity distance / 10 pc).

    .. note :: The distance modulus can be used to convert between apparent
               and absolute magnitudes via m = M + distance modulus, where m
               is the apparent magnitude and M is the absolute magnitude.

    Args:
        cosmo (:class:`~pyccl.core.Cosmology`): Cosmological parameters.
        a (float or array_like): Scale factor(s), normalized to 1 today.

    Returns:
        float or array_like: Distance modulus at a.
    """
    cosmo.compute_distances()
    return _vectorize_fn(lib.distance_modulus,
                         lib.distance_modulus_vec, cosmo, a)


def scale_factor_of_chi(cosmo, chi):
    """Scale factor, a, at a comoving radial distance chi.

    Args:
        cosmo (:class:`~pyccl.core.Cosmology`): Cosmological parameters.
        chi (float or array_like): Comoving radial distance(s); Mpc.

    Returns:
        float or array_like: Scale factor(s), normalized to 1 today.
    """
    cosmo.compute_distances()
    return _vectorize_fn(lib.scale_factor_of_chi,
                         lib.scale_factor_of_chi_vec, cosmo, chi)


def omega_x(cosmo, a, species):
    """Density fraction of a given species at a redshift different than z=0.

    Args:
        cosmo (:class:`~pyccl.core.Cosmology`): Cosmological parameters.
        a (float or array_like): Scale factor(s), normalized to 1 today.
        species (string): species type. Should be one of

            - 'matter': cold dark matter, massive neutrinos, and baryons
            - 'dark_energy': cosmological constant or otherwise
            - 'radiation': relativistic species besides massless neutrinos
            - 'curvature': curvature density
            - 'neutrinos_rel': relativistic neutrinos
            - 'neutrinos_massive': massive neutrinos

    Returns:
        float or array_like: Density fraction of a given species at a \
                             scale factor.
    """
    if species not in species_types.keys():
        raise ValueError("'%s' is not a valid species type. "
                         "Available options are: %s"
                         % (species, species_types.keys()))

    return _vectorize_fn3(lib.omega_x,
                          lib.omega_x_vec, cosmo, a, species_types[species])


def rho_x(cosmo, a, species, is_comoving=False):
    """Physical or comoving density as a function of scale factor.

    Args:
        cosmo (:class:`~pyccl.core.Cosmology`): Cosmological parameters.
        a (float or array_like): Scale factor(s), normalized to 1 today.
        species (string): species type. Should be one of

            - 'matter': cold dark matter, massive neutrinos, and baryons
            - 'dark_energy': cosmological constant or otherwise
            - 'radiation': relativistic species besides massless neutrinos
            - 'curvature': curvature density
            - 'neutrinos_rel': relativistic neutrinos
            - 'neutrinos_massive': massive neutrinos

        is_comoving (bool): either physical (False, default) or comoving (True)

    Returns:
        rho_x (float or array_like): Physical density of a given species
        at a scale factor, in units of Msun / Mpc^3.
    """

    if species not in species_types.keys():
        raise ValueError("'%s' is not a valid species type. "
                         "Available options are: %s"
                         % (species, species_types.keys()))

    return _vectorize_fn4(
        lib.rho_x, lib.rho_x_vec, cosmo, a,
        species_types[species], int(is_comoving))


def sigma_critical(cosmo, a_lens, a_source):
    """Returns the critical surface mass density.

    .. math::
         \\Sigma_{\\mathrm{crit}} = \\frac{c^2}{4\\pi G}
          \\frac{D_{\\rm{s}}}{D_{\\rm{l}}D_{\\rm{ls}}},
           where :math:`c` is the speed of light, :math:`G` is the
           gravitational constant, and :math:`D_i` is the angular diameter
           distance. The labels :math:`i =` s, l and ls denotes the distances
           to the source, lens, and between source and lens, respectively.

    Args:
        cosmo (:class:`~pyccl.core.Cosmology`): A Cosmology object.
        a_lens (float): lens' scale factor.
        a_source (float or array_like): source's scale factor.

    Returns:
        float or array_like: :math:`\\Sigma_{\\mathrm{crit}}` in units
        of :math:`\\M_{\\odot}/Mpc^2`
    """
    Ds = angular_diameter_distance(cosmo, a_source, a2=None)
    Dl = angular_diameter_distance(cosmo, a_lens, a2=None)
    Dls = angular_diameter_distance(cosmo, a_lens, a_source)
    A = (
        physical_constants.CLIGHT**2
        * physical_constants.MPC_TO_METER
        / (4.0 * np.pi * physical_constants.GNEWT
           * physical_constants.SOLAR_MASS)
    )

    Sigma_crit = A * Ds / (Dl * Dls)
    return Sigma_crit



from . import halos as hal
from .pyutils import deprecated


@deprecated(hal.HaloProfileNFW)
def nfw_profile_3d(cosmo, concentration, halo_mass, odelta, a, r):
    """Calculate the 3D NFW halo profile at a given radius or an array
    of radii, for a halo with a given mass, mass definition, and
    concentration, at a given scale factor, with a cosmology dependence.

    .. note:: Note that this function is deprecated. Please use the
              functionality in the :mod:`~pyccl.halos.profiles` module.

    Args:
        cosmo (:class:`~pyccl.core.Cosmology`): cosmological parameters.
        concentration (float): halo concentration.
        halo_mass (float): halo masses; in units of Msun.
        odelta (float): overdensity with respect to mean matter density.
        a (float): scale factor.
        r (float or array_like): radius or radii to calculate profile for,
         in units of Mpc.

    Returns:
        float or array_like: 3D NFW density at r, in units of Msun/Mpc^3.
    """
    mdef = hal.MassDef(odelta, 'matter')
    c = hal.ConcentrationConstant(c=concentration,
                                  mdef=mdef)
    p = hal.HaloProfileNFW(c, truncated=False)
    return p.real(cosmo, r, halo_mass, a, mdef)


@deprecated(hal.HaloProfileEinasto)
def einasto_profile_3d(cosmo, concentration, halo_mass, odelta, a, r):
    """Calculate the 3D Einasto halo profile
    at a given radius or an array of radii,
    for a halo with a given mass, mass definition, and concentration,
    at a given scale factor, with a cosmology dependence.
    The alpha parameter is calibrated using the relation with peak height in
    https://arxiv.org/pdf/1401.1216.pdf eqn5, assuming virial mass.

    .. note:: Note that this function is deprecated. Please use the
              functionality in the :mod:`~pyccl.halos.profiles` module.

    Args:
        cosmo (:class:`~pyccl.core.Cosmology`): cosmological parameters.
        concentration (float): halo concentration.
        halo_mass (float): halo masses; in units of Msun.
        odelta (float): overdensity with respect to mean matter density.
        a (float): scale factor.
        r (float or array_like): radius or radii to calculate profile for,
         in units of Mpc.

    Returns:
        float or array_like: 3D NFW density at r, in units of Msun/Mpc^3.
    """
    mdef = hal.MassDef(odelta, 'matter')
    c = hal.ConcentrationConstant(c=concentration,
                                  mdef=mdef)
    mdef = hal.MassDef(odelta, 'matter',
                       c_m_relation=c)
    p = hal.HaloProfileEinasto(c, truncated=False)
    return p.real(cosmo, r, halo_mass, a, mdef)


@deprecated(hal.HaloProfileHernquist)
def hernquist_profile_3d(cosmo, concentration, halo_mass, odelta, a, r):
    """Calculate the 3D Hernquist halo profile
    at a given radius or an array of radii,
    for a halo with a given mass, mass definition, and concentration,
    at a given scale factor, with a cosmology dependence.

    .. note:: Note that this function is deprecated. Please use the
              functionality in the :mod:`~pyccl.halos.profiles` module.

    Args:
        cosmo (:class:`~pyccl.core.Cosmology`): cosmological parameters.
        concentration (float): halo concentration.
        halo_mass (float): halo masses; in units of Msun.
        odelta (float): overdensity with respect to mean matter density.
        a (float): scale factor.
        r (float or array_like): radius or radii to calculate profile for,
         in units of Mpc.

    Returns:
        float or array_like: 3D NFW density at r, in units of Msun/Mpc^3.
    """
    mdef = hal.MassDef(odelta, 'matter')
    c = hal.ConcentrationConstant(c=concentration,
                                  mdef=mdef)
    p = hal.HaloProfileHernquist(c, truncated=False)
    return p.real(cosmo, r, halo_mass, a, mdef)


@deprecated(hal.HaloProfileNFW)
def nfw_profile_2d(cosmo, concentration, halo_mass, odelta, a, r):
    """Calculate the 2D projected NFW halo profile
    at a given radius or an array of radii,
    for a halo with a given mass, mass definition, and concentration,
    at a given scale factor, with a cosmology dependence.

    .. note:: Note that this function is deprecated. Please use the
              functionality in the :mod:`~pyccl.halos.profiles` module.

    Args:
        cosmo (:class:`~pyccl.core.Cosmology`): cosmological parameters.
        concentration (float): halo concentration.
        halo_mass (float): halo masses; in units of Msun.
        odelta (float): overdensity with respect to mean matter density.
        a (float): scale factor.
        r (float or array_like): radius or radii to calculate profile for,
         in units of Mpc.

    Returns:
        float or array_like: 2D projected NFW density at r, \
         in units of Msun/Mpc^2.
    """
    mdef = hal.MassDef(odelta, 'matter')
    c = hal.ConcentrationConstant(c=concentration,
                                  mdef=mdef)
    p = hal.HaloProfileNFW(c, truncated=False,
                           projected_analytic=True)
    return p.projected(cosmo, r, halo_mass, a, mdef)



import numpy as np
import pyccl as ccl
import pytest


@pytest.mark.parametrize('func', [
    'nfw_profile_3d', 'nfw_profile_2d',
    'einasto_profile_3d', 'hernquist_profile_3d'])
@pytest.mark.parametrize('r', [
    1,
    1.,
    np.array([1, 2, 3]),
    [1, 2, 3]])
def test_haloprofile_smoke(func, r):
    from pyccl.pyutils import assert_warns
    cosmo = ccl.Cosmology(
        Omega_c=0.27, Omega_b=0.045, h=0.67, sigma8=0.8, n_s=0.96,
        transfer_function='bbks', matter_power_spectrum='linear')
    a = 1
    c = 5
    mass = 1e14
    odelta = 200
    # These are all deprecated
    prof = assert_warns(
        ccl.CCLWarning,
        getattr(ccl, func), cosmo, c, mass, odelta, a, r)
    assert np.all(np.isfinite(prof))
    assert np.shape(prof) == np.shape(r)



import copy
import numpy as np
from numpy.testing import assert_raises
import pyccl
from pyccl import ccllib
from pyccl import CCLError

PYCOSMO = pyccl.Cosmology(Omega_c=0.27, Omega_b=0.045, h=0.67,
                          sigma8=0.8, n_s=0.96,
                          transfer_function='bbks')
PYCOSMO.compute_nonlin_power()
COSMO = PYCOSMO.cosmo


def test_swig_tracer():
    z = np.linspace(0, 1, 10)
    bias_ia = z * 2
    dNdz = z * 8

    assert_raises(
        ValueError,
        pyccl.WeakLensingTracer,
        PYCOSMO,
        dndz=(z, dNdz[0:2]),
        ia_bias=(z, bias_ia))

    assert_raises(
        ValueError,
        pyccl.WeakLensingTracer,
        PYCOSMO,
        dndz=(z, dNdz),
        ia_bias=(z, bias_ia[0:2]))


def test_swig_background():
    status = 0
    for func in [
            ccllib.growth_factor_vec,
            ccllib.growth_factor_unnorm_vec,
            ccllib.growth_rate_vec,
            ccllib.comoving_radial_distance_vec,
            ccllib.comoving_angular_distance_vec,
            ccllib.h_over_h0_vec,
            ccllib.luminosity_distance_vec,
            ccllib.distance_modulus_vec]:
        assert_raises(
            CCLError,
            func,
            COSMO,
            [0.0, 1.0],
            1,
            status)

    assert_raises(
        CCLError,
        ccllib.omega_x_vec,
        COSMO,
        0,
        [0.0, 1.0],
        1,
        status)

    assert_raises(
        CCLError,
        ccllib.rho_x_vec,
        COSMO,
        0,
        0,
        [0.0, 1.0],
        1,
        status)

    assert_raises(
        CCLError,
        ccllib.scale_factor_of_chi_vec,
        COSMO,
        [0.0, 1.0],
        1,
        status)


def test_swig_cls():
    status = 0
    base_args = [
        [0.0, 1.0],
        [0.0, 1.0],
        [0.0, 1.0, 2.0, 3.0],
        [0.0, 1.0, 2.0],
        [0.0],
        [0.0, 1.0, 2.0],
        [0.0, 1.0, 2.0, 3.0],
        0, 1, 0, 0, 0, 0, 2,
        status]
    args = copy.deepcopy(base_args)
    args[1] = [1.0] * 8
    assert_raises(
        CCLError,
        ccllib.cl_tracer_t_new_wrapper,
        COSMO, 0, 0,
        *args)

    assert_raises(
        CCLError,
        ccllib.angular_cl_vec,
        COSMO,
        None, None, None,
        1, 0, pyccl.pyutils.integ_types['spline'],
        "none",
        status)


def test_swig_core():
    status = 0
    assert_raises(
        CCLError,
        ccllib.parameters_create_nu_vec,
        0.25, 0.05, 0.0, 3.0, -1.0, 0.0, 0.7, 2e-9, 0.95, 1, 0.0, 0.0,
        0.0, 0.0, 1.0, 1.0, 0.0, [1.0, 2.0],
        [0.0, 0.3, 0.5],
        [0.02, 0.01, 0.2],
        status)


def test_swig_correlation():
    status = 0
    assert_raises(
        CCLError,
        ccllib.correlation_vec,
        COSMO,
        [1, 2],
        [1, 2, 3],
        [0, 1, 2, 3],
        0, 0,
        4,
        status)

    assert_raises(
        CCLError,
        ccllib.correlation_vec,
        COSMO,
        [1, 2, 3],
        [1, 2, 3],
        [0, 1, 2, 3],
        0, 0,
        5,
        status)

    assert_raises(
        CCLError,
        ccllib.correlation_3d_vec,
        COSMO,
        PYCOSMO._pk_nl['delta_matter:delta_matter'].psp,
        1.0,
        [1, 2, 3],
        4,
        status)


def test_swig_neurtinos():
    status = 0
    assert_raises(
        CCLError,
        ccllib.Omeganuh2_vec,
        3, 2.7,
        [0.0, 1.0],
        [0.05, 0.1, 0.2],
        4,
        status)

    assert_raises(
        CCLError,
        ccllib.Omeganuh2_vec,
        3, 2.7,
        [0.0, 1.0],
        [0.1, 0.2],
        2,
        status)


def test_swig_power():
    status = 0

    for func in [ccllib.sigmaR_vec,
                 ccllib.sigmaV_vec]:
        assert_raises(
            CCLError,
            func,
            COSMO,
            None,
            1.0,
            [1.0, 2.0],
            3,
            status)

    assert_raises(
        CCLError,
        ccllib.kNL_vec,
        COSMO,
        None,
        [0.5, 1.0],
        3,
        status)



# flake8: noqa E402
from pkg_resources import get_distribution, DistributionNotFound
try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    # package is not installed
    pass
del get_distribution, DistributionNotFound

# Set the environment variable for default config path
from os import environ, path
if environ.get("CLASS_PARAM_DIR") is None:
    environ["CLASS_PARAM_DIR"] = path.dirname(path.abspath(__file__))
del environ, path

# SWIG-generated
from . import ccllib as lib

# Errors
from .errors import (
    CCLError,
    CCLWarning,
    CCLDeprecationWarning,
)

# Constants and accuracy parameters
from .parameters import (
    CCLParameters,
    gsl_params,
    spline_params,
    physical_constants,
)

# Core data structures
from .core import (
    Cosmology,
    CosmologyVanillaLCDM,
    CosmologyCalculator,
)

# Background cosmology functions and growth functions
from .background import (
    growth_factor,
    growth_factor_unnorm,
    growth_rate,
    comoving_radial_distance,
    angular_diameter_distance,
    comoving_angular_distance,
    luminosity_distance,
    distance_modulus,
    h_over_h0,
    scale_factor_of_chi,
    omega_x,
    rho_x,
    sigma_critical,
)

# Boltzmann solvers
from .boltzmann import (
    get_camb_pk_lin,
    get_isitgr_pk_lin,
    get_class_pk_lin,
)

# Generalized power spectra
from .pk2d import (
    Pk2D,
    parse_pk2d,
)

# Generalized connected trispectra
from .tk3d import Tk3D

# Power spectrum calculations, sigma8 and kNL
from .power import (
    linear_power,
    nonlin_power,
    linear_matter_power,
    nonlin_matter_power,
    sigmaR,
    sigmaV,
    sigma8,
    sigmaM,
    kNL,
)

# Baryons & Neutrinos
from .bcm import (
    bcm_model_fka,
    bcm_correct_pk2d,
)

from .neutrinos import (
    Omeganuh2,
    nu_masses,
)

# Cells & Tracers
from .cls import angular_cl
from .tracers import (
    Tracer,
    NumberCountsTracer,
    WeakLensingTracer,
    CMBLensingTracer,
    tSZTracer,
    CIBTracer,
    ISWTracer,
    get_density_kernel,
    get_kappa_kernel,
    get_lensing_kernel,
)

# Correlations & Covariances
from .correlations import (
    correlation,
    correlation_3d,
    correlation_multipole,
    correlation_3dRsd,
    correlation_3dRsd_avgmu,
    correlation_pi_sigma,
)

from .covariances import (
    angular_cl_cov_cNG,
    angular_cl_cov_SSC,
    sigma2_B_disc,
    sigma2_B_from_mask,
)


# Miscellaneous
from .pyutils import debug_mode, resample_array


# Deprecated & Renamed modules
from .halomodel import (
    halomodel_matter_power,
    halo_concentration,
    onehalo_matter_power,
    twohalo_matter_power,
)

from .massfunction import (
    massfunc,
    halo_bias,
    massfunc_m2r,
)

from .haloprofile import (
    nfw_profile_3d,
    einasto_profile_3d,
    hernquist_profile_3d,
    nfw_profile_2d,
)


__all__ = (
    'lib',
    'CCLParameters', 'spline_params', 'gsl_params', 'physical_constants',
    'CCLError', 'CCLWarning', 'CCLDeprecationWarning',
    'Cosmology', 'CosmologyVanillaLCDM', 'CosmologyCalculator',
    'growth_factor', 'growth_factor_unnorm', 'growth_rate',
    'comoving_radial_distance', 'angular_diameter_distance',
    'comoving_angular_distance', 'luminosity_distance', 'distance_modulus',
    'h_over_h0', 'scale_factor_of_chi', 'omega_x', 'rho_x', 'sigma_critical',
    'get_camb_pk_lin', 'get_isitgr_pk_lin', 'get_class_pk_lin',
    'Pk2D', 'parse_pk2d', 'Tk3D',
    'linear_power', 'nonlin_power',
    'linear_matter_power', 'nonlin_matter_power',
    'sigmaR', 'sigmaV', 'sigma8', 'sigmaM', 'kNL',
    'bcm_model_fka', 'bcm_correct_pk2d',
    'Omeganuh2', 'nu_masses',
    'angular_cl',
    'Tracer', 'NumberCountsTracer', 'WeakLensingTracer', 'CMBLensingTracer',
    'tSZTracer', 'CIBTracer', 'ISWTracer',
    'get_density_kernel', 'get_kappa_kernel', 'get_lensing_kernel',
    'correlation', 'correlation_3d', 'correlation_multipole',
    'correlation_3dRsd', 'correlation_3dRsd_avgmu', 'correlation_pi_sigma',
    'angular_cl_cov_cNG', 'angular_cl_cov_SSC',
    'sigma2_B_disc', 'sigma2_B_from_mask',
    'debug_mode', 'resample_array',
    'halomodel_matter_power', 'halo_concentration',
    'onehalo_matter_power', 'twohalo_matter_power',
    'massfunc', 'halo_bias', 'massfunc_m2r', 'nfw_profile_3d',
    'einasto_profile_3d', 'hernquist_profile_3d', 'nfw_profile_2d',
)



from .. import ccllib as lib
from ..core import check
from ..background import omega_x
from .massdef import MassDef, MassDef200m
import numpy as np
import functools


class HaloBias(object):
    """ This class enables the calculation of halo bias functions.
    We currently assume that all halo bias functions can be written
    as functions that depend on M only through sigma_M (where
    sigma_M^2 is the overdensity variance on spheres with a
    radius given by the Lagrangian radius for mass M).
    All sub-classes implementing specific parametrizations
    can therefore be simply created by replacing this class'
    `_get_bsigma method`.

    Args:
        cosmo (:class:`~pyccl.core.Cosmology`): A Cosmology object.
        mass_def (:class:`~pyccl.halos.massdef.MassDef`): a mass
            definition object that fixes
            the mass definition used by this halo bias
            parametrization.
        mass_def_strict (bool): if False, consistency of the mass
            definition will be ignored.
    """
    name = "default"

    def __init__(self, cosmo, mass_def=None, mass_def_strict=True):
        cosmo.compute_sigma()
        self.mass_def_strict = mass_def_strict
        if mass_def is not None:
            if self._check_mdef(mass_def):
                raise ValueError("Halo bias " + self.name +
                                 " is not compatible with mass definition" +
                                 " Delta = %s, " % (mass_def.Delta) +
                                 " rho = " + mass_def.rho_type)
            self.mdef = mass_def
        else:
            self._default_mdef()
        self._setup(cosmo)

    def _default_mdef(self):
        """ Assigns a default mass definition for this object if
        none is passed at initialization.
        """
        self.mdef = MassDef('fof', 'matter')

    def _setup(self, cosmo):
        """ Use this function to initialize any internal attributes
        of this object. This function is called at the very end of the
        constructor call.

        Args:
            cosmo (:class:`~pyccl.core.Cosmology`): A Cosmology object.
        """
        pass

    def _check_mdef_strict(self, mdef):
        return False

    def _check_mdef(self, mdef):
        """ Return False if the input mass definition agrees with
        the definitions for which this parametrization
        works. True otherwise. This function gets called at the
        start of the constructor call.

        Args:
            mdef (:class:`~pyccl.halos.massdef.MassDef`):
                a mass definition object.

        Returns:
            bool: True if the mass definition is not compatible with
                this parametrization. False otherwise.
        """
        if self.mass_def_strict:
            return self._check_mdef_strict(mdef)
        return False

    def _get_consistent_mass(self, cosmo, M, a, mdef_other):
        """ Transform a halo mass with a given mass definition into
        the corresponding mass definition that was used to initialize
        this object.

        Args:
            cosmo (:class:`~pyccl.core.Cosmology`): A Cosmology object.
            M (float or array_like): halo mass in units of M_sun.
            a (float): scale factor.
            mdef_other (:class:`~pyccl.halos.massdef.MassDef`):
                a mass definition object.

        Returns:
            float or array_like: mass according to this object's
            mass definition.
        """
        if mdef_other is not None:
            M_use = mdef_other.translate_mass(cosmo, M, a, self.mdef)
        else:
            M_use = M
        return np.log10(M_use)

    def _get_Delta_m(self, cosmo, a):
        """ For SO-based mass definitions, this returns the corresponding
        value of Delta for a rho_matter-based definition. This is useful
        mostly for the Tinker mass functions, which are defined for any
        SO mass in general, but explicitly only for Delta_matter.
        """
        delta = self.mdef.get_Delta(cosmo, a)
        if self.mdef.rho_type == 'matter':
            return delta
        else:
            om_this = omega_x(cosmo, a, self.mdef.rho_type)
            om_matt = omega_x(cosmo, a, 'matter')
            return delta * om_this / om_matt

    def get_halo_bias(self, cosmo, M, a, mdef_other=None):
        """ Returns the halo bias for input parameters.

        Args:
            cosmo (:class:`~pyccl.core.Cosmology`): A Cosmology object.
            M (float or array_like): halo mass in units of M_sun.
            a (float): scale factor.
            mdef_other (:class:`~pyccl.halos.massdef.MassDef`):
                the mass definition object that defines M.

        Returns:
            float or array_like: halo bias.
        """
        M_use = np.atleast_1d(M)
        logM = self._get_consistent_mass(cosmo, M_use,
                                         a, mdef_other)

        # sigma(M)
        status = 0
        sigM, status = lib.sigM_vec(cosmo.cosmo, a, logM,
                                    len(logM), status)
        check(status, cosmo=cosmo)

        b = self._get_bsigma(cosmo, sigM, a)
        if np.ndim(M) == 0:
            b = b[0]
        return b

    def _get_bsigma(self, cosmo, sigM, a):
        """ Get the halo bias as a function of sigmaM.

        Args:
            cosmo (:class:`~pyccl.core.Cosmology`): A Cosmology object.
            sigM (float or array_like): standard deviation in the
                overdensity field on the scale of this halo.
            a (float): scale factor.

        Returns:
            float or array_like: f(sigma_M) function.
        """
        raise NotImplementedError("Use one of the non-default "
                                  "HaloBias classes")

    @classmethod
    def from_name(cls, name):
        """Returns halo bias subclass from name string

        Args:
            name (string): a halo bias name

        Returns:
            HaloBias subclass corresponding to the input name.
        """
        bias_functions = {c.name: c for c in HaloBias.__subclasses__()}
        if name in bias_functions:
            return bias_functions[name]
        else:
            raise ValueError(
                f"Halo bias parametrization {name} not implemented")


class HaloBiasSheth99(HaloBias):
    """ Implements halo bias described in 1999MNRAS.308..119S
    This parametrization is only valid for 'fof' masses.

    Args:
        cosmo (:class:`~pyccl.core.Cosmology`): A Cosmology object.
        mass_def (:class:`~pyccl.halos.massdef.MassDef`):
            a mass definition object.
            this parametrization accepts FoF masses only.
            If `None`, FoF masses will be used.
        mass_def_strict (bool): if False, consistency of the mass
            definition will be ignored.
        use_delta_c_fit (bool): if True, use delta_crit given by
            the fit of Nakamura & Suto 1997. Otherwise use
            delta_crit = 1.68647.
    """
    name = "Sheth99"

    def __init__(self, cosmo, mass_def=None,
                 mass_def_strict=True,
                 use_delta_c_fit=False):
        self.use_delta_c_fit = use_delta_c_fit
        super(HaloBiasSheth99, self).__init__(cosmo,
                                              mass_def,
                                              mass_def_strict)

    def _default_mdef(self):
        self.mdef = MassDef('fof', 'matter')

    def _setup(self, cosmo):
        self.p = 0.3
        self.a = 0.707

    def _check_mdef_strict(self, mdef):
        if self.mass_def_strict:
            if mdef.Delta != 'fof':
                return True
        return False

    def _get_bsigma(self, cosmo, sigM, a):
        if self.use_delta_c_fit:
            status = 0
            delta_c, status = lib.dc_NakamuraSuto(cosmo.cosmo, a, status)
            check(status, cosmo=cosmo)
        else:
            delta_c = 1.68647

        nu = delta_c / sigM
        anu2 = self.a * nu**2
        return 1. + (anu2 - 1. + 2. * self.p / (1. + anu2**self.p))/delta_c


class HaloBiasSheth01(HaloBias):
    """ Implements halo bias described in arXiv:astro-ph/9907024.
    This parametrization is only valid for 'fof' masses.

    Args:
        cosmo (:class:`~pyccl.core.Cosmology`): A Cosmology object.
        mass_def (:class:`~pyccl.halos.massdef.MassDef`):
            a mass definition object.
            this parametrization accepts FoF masses only.
            If `None`, FoF masses will be used.
        mass_def_strict (bool): if False, consistency of the mass
            definition will be ignored.
    """
    name = "Sheth01"

    def __init__(self, cosmo, mass_def=None, mass_def_strict=True):
        super(HaloBiasSheth01, self).__init__(cosmo,
                                              mass_def,
                                              mass_def_strict)

    def _default_mdef(self):
        self.mdef = MassDef('fof', 'matter')

    def _setup(self, cosmo):
        self.a = 0.707
        self.sqrta = 0.84083292038
        self.b = 0.5
        self.c = 0.6
        self.dc = 1.68647

    def _check_mdef_strict(self, mdef):
        if mdef.Delta != 'fof':
            return True
        return False

    def _get_bsigma(self, cosmo, sigM, a):
        nu = self.dc/sigM
        anu2 = self.a * nu**2
        anu2c = anu2**self.c
        t1 = self.b * (1.0 - self.c) * (1.0 - 0.5 * self.c)
        return 1. + (self.sqrta * anu2 * (1 + self.b / anu2c) -
                     anu2c / (anu2c + t1)) / (self.sqrta * self.dc)


class HaloBiasBhattacharya11(HaloBias):
    """ Implements halo bias described in arXiv:1005.2239.
    This parametrization is only valid for 'fof' masses.

    Args:
        cosmo (:class:`~pyccl.core.Cosmology`): A Cosmology object.
        mass_def (:class:`~pyccl.halos.massdef.MassDef`):
            a mass definition object.
            this parametrization accepts FoF masses only.
            If `None`, FoF masses will be used.
        mass_def_strict (bool): if False, consistency of the mass
            definition will be ignored.
    """
    name = "Bhattacharya11"

    def __init__(self, cosmo, mass_def=None, mass_def_strict=True):
        super(HaloBiasBhattacharya11, self).__init__(cosmo,
                                                     mass_def,
                                                     mass_def_strict)

    def _default_mdef(self):
        self.mdef = MassDef('fof', 'matter')

    def _setup(self, cosmo):
        self.a = 0.788
        self.az = 0.01
        self.p = 0.807
        self.q = 1.795
        self.dc = 1.68647

    def _check_mdef_strict(self, mdef):
        if mdef.Delta != 'fof':
            return True
        return False

    def _get_bsigma(self, cosmo, sigM, a):
        nu = self.dc / sigM
        a = self.a * a**self.az
        anu2 = a * nu**2
        return 1. + (anu2 - self.q + 2*self.p / (1 + anu2**self.p)) / self.dc


class HaloBiasTinker10(HaloBias):
    """ Implements halo bias described in arXiv:1001.3162.

    Args:
        cosmo (:class:`~pyccl.core.Cosmology`): A Cosmology object.
        mass_def (:class:`~pyccl.halos.massdef.MassDef`):
            a mass definition object.
            this parametrization accepts SO masses with
            200 < Delta < 3200 with respect to the matter density.
            If `None`, Delta = 200 (matter) will be used.
        mass_def_strict (bool): if False, consistency of the mass
            definition will be ignored.
    """
    name = "Tinker10"

    def __init__(self, cosmo, mass_def=None, mass_def_strict=True):
        super(HaloBiasTinker10, self).__init__(cosmo,
                                               mass_def,
                                               mass_def_strict)

    def _default_mdef(self):
        self.mdef = MassDef200m()

    def _AC(self, ld):
        xp = np.exp(-(4./ld)**4.)
        A = 1.0 + 0.24 * ld * xp
        C = 0.019 + 0.107 * ld + 0.19*xp
        return A, C

    def _a(self, ld):
        return 0.44 * ld - 0.88

    def _setup(self, cosmo):
        self.B = 0.183
        self.b = 1.5
        self.c = 2.4
        self.dc = 1.68647

    def _check_mdef_strict(self, mdef):
        if mdef.Delta == 'fof':
            return True
        return False

    def _get_bsigma(self, cosmo, sigM, a):
        nu = self.dc / sigM

        ld = np.log10(self._get_Delta_m(cosmo, a))
        A, C = self._AC(ld)
        aa = self._a(ld)
        nupa = nu**aa
        return 1. - A * nupa / (nupa + self.dc**aa) + \
            self.B * nu**self.b + C * nu**self.c


@functools.wraps(HaloBias.from_name)
def halo_bias_from_name(name):
    return HaloBias.from_name(name)



import numpy as np
import pytest
import pyccl as ccl


COSMO = ccl.Cosmology(
    Omega_c=0.27, Omega_b=0.045, h=0.67, sigma8=0.8, n_s=0.96,
    transfer_function='bbks', matter_power_spectrum='halofit')
MS = [1E13, [1E12, 1E15], np.array([1E12, 1E15])]
MF_EQUIV = {'tinker10': 'Tinker10',
            'tinker': 'Tinker08',
            'watson': 'Watson13',
            'shethtormen': 'Sheth99',
            'angulo': 'Angulo12'}
MF_TYPES = sorted(list(MF_EQUIV.keys()))


@pytest.mark.parametrize('mf_type', MF_TYPES)
def test_massfunc_models_smoke(mf_type):
    from pyccl.pyutils import assert_warns
    cosmo = ccl.Cosmology(
        Omega_c=0.27, Omega_b=0.045, h=0.67, sigma8=0.8, n_s=0.96,
        transfer_function='bbks', matter_power_spectrum='linear',
        mass_function=mf_type)
    hmf_cls = ccl.halos.mass_function_from_name(MF_EQUIV[mf_type])
    hmf = hmf_cls(cosmo)
    for m in MS:
        # Deprecated
        nm_old = assert_warns(ccl.CCLWarning, ccl.massfunc, cosmo, m, 1.)
        nm_new = hmf.get_mass_function(cosmo, m, 1.)
        assert np.all(np.isfinite(nm_old))
        assert np.shape(nm_old) == np.shape(m)
        assert np.all(np.array(nm_old) ==
                      np.array(nm_new))


@pytest.mark.parametrize('mf_type', ['tinker10', 'shethtormen'])
def test_halo_bias_models_smoke(mf_type):
    from pyccl.pyutils import assert_warns
    cosmo = ccl.Cosmology(
        Omega_c=0.27, Omega_b=0.045, h=0.67, sigma8=0.8, n_s=0.96,
        transfer_function='bbks', matter_power_spectrum='linear',
        mass_function=mf_type)
    hbf_cls = ccl.halos.halo_bias_from_name(MF_EQUIV[mf_type])
    hbf = hbf_cls(cosmo)
    for m in MS:
        # Deprecated
        bm_old = assert_warns(ccl.CCLWarning, ccl.halo_bias, cosmo, m, 1.)
        bm_new = hbf.get_halo_bias(cosmo, m, 1.)
        assert np.all(np.isfinite(bm_old))
        assert np.shape(bm_old) == np.shape(m)
        assert np.all(np.array(bm_old) ==
                      np.array(bm_new))


@pytest.mark.parametrize('m', [
    1e14,
    int(1e14),
    [1e14, 1e15],
    np.array([1e14, 1e15])])
def test_massfunc_smoke(m):
    a = 0.8
    mf = ccl.halos.MassFuncTinker10(COSMO).get_mass_function(COSMO, m, a)
    assert np.all(np.isfinite(mf))
    assert np.shape(mf) == np.shape(m)


@pytest.mark.parametrize('m', [
    1e14,
    int(1e14),
    [1e14, 1e15],
    np.array([1e14, 1e15])])
def test_massfunc_m2r_smoke(m):
    from pyccl.pyutils import assert_warns
    # Deprecated
    # TODO: switch to mass2radius_lagrangian
    r = assert_warns(ccl.CCLWarning, ccl.massfunc_m2r, COSMO, m)
    assert np.all(np.isfinite(r))
    assert np.shape(r) == np.shape(m)


@pytest.mark.parametrize('m', [
    1e14,
    int(1e14),
    [1e14, 1e15],
    np.array([1e14, 1e15])])
def test_sigmaM_smoke(m):
    a = 0.8
    s = ccl.sigmaM(COSMO, m, a)
    assert np.all(np.isfinite(s))
    assert np.shape(s) == np.shape(m)


@pytest.mark.parametrize('m', [
    1e14,
    int(1e14),
    [1e14, 1e15],
    np.array([1e14, 1e15])])
def test_halo_bias_smoke(m):
    from pyccl.pyutils import assert_warns
    a = 0.8
    # Deprecated
    # TODO: switch to HaloBias
    b = assert_warns(ccl.CCLWarning, ccl.halo_bias, COSMO, m, a)
    assert np.all(np.isfinite(b))
    assert np.shape(b) == np.shape(m)



from . import ccllib as lib
from .pyutils import check
from .pk2d import Pk2D
import numpy as np


def bcm_model_fka(cosmo, k, a):
    """The BCM model correction factor for baryons.

    .. note:: BCM stands for the "baryonic correction model" of Schneider &
              Teyssier (2015; https://arxiv.org/abs/1510.06034). See the
              `DESC Note <https://github.com/LSSTDESC/CCL/blob/master/doc\
/0000-ccl_note/main.pdf>`_
              for details.

    .. note:: The correction factor is applied multiplicatively so that
              :math:`P_{\\rm corrected}(k, a) = P(k, a)\\, f_{\\rm bcm}(k, a)`.

    Args:
        cosmo (:class:`~pyccl.core.Cosmology`): Cosmological parameters.
        k (float or array_like): Wavenumber; Mpc^-1.
        a (float): Scale factor.

    Returns:
        float or array_like: Correction factor to apply to the power spectrum.
    """
    k_use = np.atleast_1d(k)
    status = 0
    fka, status = lib.bcm_model_fka_vec(cosmo.cosmo, a, k_use,
                                        len(k_use), status)
    check(status, cosmo)

    if np.ndim(k) == 0:
        fka = fka[0]
    return fka


def bcm_correct_pk2d(cosmo, pk2d):
    """Apply the BCM model correction factor to a given power spectrum.

    Args:
        cosmo (:class:`~pyccl.core.Cosmology`): Cosmological parameters.
        pk2d (:class:`~pyccl.pk2d.Pk2D`): power spectrum.
    """
    if not isinstance(pk2d, Pk2D):
        raise TypeError("pk2d must be a Pk2D object")
    status = 0
    status = lib.bcm_correct(cosmo.cosmo, pk2d.psp, status)
    check(status, cosmo)



import numpy as np
import pytest
import pyccl as ccl


COSMO = ccl.Cosmology(
    Omega_c=0.27, Omega_b=0.045, h=0.67, sigma8=0.8, n_s=0.96,
    transfer_function='bbks', matter_power_spectrum='linear')
M200 = ccl.halos.MassDef200m()
HMF = ccl.halos.MassFuncTinker10(COSMO, mass_def=M200)
HBF = ccl.halos.HaloBiasTinker10(COSMO, mass_def=M200)
P1 = ccl.halos.HaloProfileNFW(ccl.halos.ConcentrationDuffy08(M200),
                              fourier_analytic=True)
P2 = ccl.halos.HaloProfileHOD(ccl.halos.ConcentrationDuffy08(M200))
P3 = ccl.halos.HaloProfilePressureGNFW()
P4 = P1
Pneg = ccl.halos.HaloProfilePressureGNFW(P0=-1)
PKC = ccl.halos.Profile2pt()
PKCH = ccl.halos.Profile2ptHOD()
KK = np.geomspace(1E-3, 10, 32)
MM = np.geomspace(1E11, 1E15, 16)
AA = 1.0


def smoke_assert_tkk1h_real(func):
    sizes = [(0, 0),
             (2, 0),
             (0, 2),
             (2, 3),
             (1, 3),
             (3, 1)]
    shapes = [(),
              (2,),
              (2, 2,),
              (2, 3, 3),
              (1, 3, 3),
              (3, 1, 1)]
    for (sa, sk), sh in zip(sizes, shapes):
        if sk == 0:
            k = 0.1
        else:
            k = np.logspace(-2., 0., sk)
        if sa == 0:
            a = 1.
        else:
            a = np.linspace(0.5, 1., sa)
        p = func(k, a)
        assert np.shape(p) == sh
        assert np.all(np.isfinite(p))


@pytest.mark.parametrize('pars',
                         [{'p1': P1, 'p2': None, 'cv12': None,
                           'p3': None, 'p4': None, 'cv34': None,
                           'norm': False},
                          {'p1': P1, 'p2': None, 'cv12': None,
                           'p3': None, 'p4': None, 'cv34': None,
                           'norm': True},
                          {'p1': P1, 'p2': P2, 'cv12': None,
                           'p3': None, 'p4': None, 'cv34': None,
                           'norm': False},
                          {'p1': P1, 'p2': None, 'cv12': None,
                           'p3': P3, 'p4': None, 'cv34': None,
                           'norm': False},
                          {'p1': P1, 'p2': None, 'cv12': None,
                           'p3': None, 'p4': P4, 'cv34': None,
                           'norm': False},
                          {'p1': P1, 'p2': P2, 'cv12': None,
                           'p3': P3, 'p4': P4, 'cv34': None,
                           'norm': False},
                          {'p1': P2, 'p2': P2, 'cv12': PKCH,
                           'p3': P2, 'p4': P2, 'cv34': None,
                           'norm': False},
                          {'p1': P1, 'p2': P2, 'cv12': PKC,
                           'p3': P3, 'p4': P4, 'cv34': PKC,
                           'norm': True}],)
def test_tkk1h_smoke(pars):
    hmc = ccl.halos.HMCalculator(COSMO, HMF, HBF, mass_def=M200,
                                 nlog10M=2)

    def f(k, a):
        return ccl.halos.halomod_trispectrum_1h(COSMO, hmc, k, a,
                                                prof1=pars['p1'],
                                                prof2=pars['p2'],
                                                prof12_2pt=pars['cv12'],
                                                prof3=pars['p3'],
                                                prof4=pars['p4'],
                                                prof34_2pt=pars['cv34'],
                                                normprof1=pars['norm'],
                                                normprof2=pars['norm'],
                                                normprof3=pars['norm'],
                                                normprof4=pars['norm'])
    smoke_assert_tkk1h_real(f)


def test_tkk1h_tk3d():
    hmc = ccl.halos.HMCalculator(COSMO, HMF, HBF, mass_def=M200)
    k_arr = KK
    a_arr = np.array([0.1, 0.4, 0.7, 1.0])
    tkk_arr = ccl.halos.halomod_trispectrum_1h(COSMO, hmc, k_arr, a_arr,
                                               P1, prof2=P2,
                                               prof12_2pt=PKC,
                                               prof3=P3, prof4=P4,
                                               prof34_2pt=PKC,
                                               normprof1=True,
                                               normprof2=True,
                                               normprof3=True,
                                               normprof4=True)

    # Input sampling
    tk3d = ccl.halos.halomod_Tk3D_1h(COSMO, hmc,
                                     P1, prof2=P2,
                                     prof12_2pt=PKC,
                                     prof3=P3, prof4=P4,
                                     prof34_2pt=PKC,
                                     normprof1=True,
                                     normprof2=True,
                                     normprof3=True,
                                     normprof4=True,
                                     lk_arr=np.log(k_arr),
                                     a_arr=a_arr,
                                     use_log=True)
    tkk_arr_2 = np.array([tk3d.eval(k_arr, a) for a in a_arr])
    assert np.all(np.fabs((tkk_arr / tkk_arr_2 - 1)).flatten()
                  < 1E-4)

    # Standard sampling
    tk3d = ccl.halos.halomod_Tk3D_1h(COSMO, hmc,
                                     P1, prof2=P2,
                                     prof12_2pt=PKC,
                                     prof3=P3, prof4=P4,
                                     prof34_2pt=PKC,
                                     normprof1=True,
                                     normprof2=True,
                                     normprof3=True,
                                     normprof4=True,
                                     lk_arr=np.log(k_arr),
                                     use_log=True)
    tkk_arr_2 = np.array([tk3d.eval(k_arr, a) for a in a_arr])
    assert np.all(np.fabs((tkk_arr / tkk_arr_2 - 1)).flatten()
                  < 1E-4)


def test_tkk1h_errors():
    from pyccl.pyutils import assert_warns

    hmc = ccl.halos.HMCalculator(COSMO, HMF, HBF, mass_def=M200)
    k_arr = KK
    a_arr = np.array([0.1, 0.4, 0.7, 1.0])

    # Wrong first profile
    with pytest.raises(TypeError):
        ccl.halos.halomod_trispectrum_1h(COSMO, hmc, k_arr, a_arr, None)
    # Wrong other profiles
    with pytest.raises(TypeError):
        ccl.halos.halomod_trispectrum_1h(COSMO, hmc, k_arr, a_arr,
                                         P1, prof2=PKC)
    with pytest.raises(TypeError):
        ccl.halos.halomod_trispectrum_1h(COSMO, hmc, k_arr, a_arr,
                                         P1, prof3=PKC)
    with pytest.raises(TypeError):
        ccl.halos.halomod_trispectrum_1h(COSMO, hmc, k_arr, a_arr,
                                         P1, prof4=PKC)
    # Wrong 2pts
    with pytest.raises(TypeError):
        ccl.halos.halomod_trispectrum_1h(COSMO, hmc, k_arr, a_arr,
                                         P1, prof12_2pt=P2)
    with pytest.raises(TypeError):
        ccl.halos.halomod_trispectrum_1h(COSMO, hmc, k_arr, a_arr,
                                         P1, prof34_2pt=P2)

    # Negative profile in logspace
    assert_warns(ccl.CCLWarning, ccl.halos.halomod_Tk3D_1h,
                 COSMO, hmc, P3, prof2=Pneg,
                 lk_arr=np.log(k_arr), a_arr=a_arr,
                 use_log=True)



import itertools
import numpy as np
import pytest
import pyccl as ccl
from pyccl.halos.halo_model import halomod_bias_1pt
from pyccl.pyutils import assert_warns


COSMO = ccl.Cosmology(
    Omega_c=0.27, Omega_b=0.045, h=0.67, sigma8=0.8, n_s=0.96,
    transfer_function='bbks', matter_power_spectrum='linear')
COSMO.compute_nonlin_power()
M200 = ccl.halos.MassDef200m()
HMF = ccl.halos.MassFuncTinker10(COSMO, mass_def=M200)
HBF = ccl.halos.HaloBiasTinker10(COSMO, mass_def=M200)
P1 = ccl.halos.HaloProfileNFW(ccl.halos.ConcentrationDuffy08(M200),
                              fourier_analytic=True)
# P2 will have is_number_counts = True
P2 = ccl.halos.HaloProfileHOD(ccl.halos.ConcentrationDuffy08(M200))
P2_nogc = ccl.halos.HaloProfileHOD(ccl.halos.ConcentrationDuffy08(M200))
P2_nogc.is_number_counts = False
P3 = ccl.halos.HaloProfilePressureGNFW()
P4 = P1
Pneg = ccl.halos.HaloProfilePressureGNFW(P0=-1)
PKC = ccl.halos.Profile2pt()
PKCH = ccl.halos.Profile2ptHOD()
KK = np.geomspace(1E-3, 10, 32)
MM = np.geomspace(1E11, 1E15, 16)
AA = 1.0


def get_ssc_counterterm_gc(k, a, hmc, prof1, prof2, prof12_2pt,
                           normalize=False):

    P_12 = b1 = b2 = np.zeros_like(k)
    if prof1.is_number_counts or prof2.is_number_counts:
        norm1 = hmc.profile_norm(COSMO, a, prof1)
        norm2 = hmc.profile_norm(COSMO, a, prof2)
        norm12 = 1
        if prof1.is_number_counts or normalize:
            norm12 *= norm1
        if prof2.is_number_counts or normalize:
            norm12 *= norm2

        i11_1 = hmc.I_1_1(COSMO, k, a, prof1)
        i11_2 = hmc.I_1_1(COSMO, k, a, prof2)
        i02_12 = hmc.I_0_2(COSMO, k, a, prof1, prof12_2pt, prof2)

        pk = ccl.linear_matter_power(COSMO, k, a)
        P_12 = norm12 * (pk * i11_1 * i11_2 + i02_12)

        if prof1.is_number_counts:
            b1 = halomod_bias_1pt(COSMO, hmc, k, a, prof1) * norm1
        if prof2.is_number_counts:
            b2 = halomod_bias_1pt(COSMO, hmc, k, a, prof2) * norm2

    return (b1 + b2) * P_12


@pytest.mark.parametrize('pars',
                         [{'p1': P1, 'p2': None, 'cv12': None,
                           'p3': None, 'p4': None, 'cv34': None,
                           'norm': False, 'pk': None},
                          {'p1': P1, 'p2': None, 'cv12': None,
                           'p3': None, 'p4': None, 'cv34': None,
                           'norm': True, 'pk': None},
                          {'p1': P1, 'p2': P2, 'cv12': None,
                           'p3': None, 'p4': None, 'cv34': None,
                           'norm': True, 'pk': None},
                          {'p1': P1, 'p2': None, 'cv12': None,
                           'p3': P3, 'p4': None, 'cv34': None,
                           'norm': False, 'pk': None},
                          {'p1': P1, 'p2': None, 'cv12': None,
                           'p3': None, 'p4': P4, 'cv34': None,
                           'norm': False, 'pk': None},
                          {'p1': P1, 'p2': P2, 'cv12': None,
                           'p3': P3, 'p4': P4, 'cv34': None,
                           'norm': True, 'pk': None},
                          {'p1': P2, 'p2': P2, 'cv12': PKCH,
                           'p3': P2, 'p4': P2, 'cv34': None,
                           'norm': True, 'pk': None},
                          {'p1': P1, 'p2': P2, 'cv12': PKC,
                           'p3': P3, 'p4': P4, 'cv34': PKC,
                           'norm': True, 'pk': None},
                          {'p1': P1, 'p2': None, 'cv12': None,
                           'p3': None, 'p4': None, 'cv34': None,
                           'norm': False, 'pk': 'linear'},
                          {'p1': P1, 'p2': None, 'cv12': None,
                           'p3': None, 'p4': None, 'cv34': None,
                           'norm': False, 'pk': 'nonlinear'},
                          {'p1': P1, 'p2': None, 'cv12': None,
                           'p3': None, 'p4': None, 'cv34': None,
                           'norm': False, 'pk': COSMO.get_nonlin_power()},
                          ],)
def test_tkkssc_smoke(pars):
    hmc = ccl.halos.HMCalculator(COSMO, HMF, HBF, mass_def=M200,
                                 nlog10M=2)
    k_arr = KK
    a_arr = np.array([0.3, 0.5, 0.7, 1.0])

    tkk = ccl.halos.halomod_Tk3D_SSC(COSMO, hmc,
                                     prof1=pars['p1'],
                                     prof2=pars['p2'],
                                     prof12_2pt=pars['cv12'],
                                     prof3=pars['p3'],
                                     prof4=pars['p4'],
                                     prof34_2pt=pars['cv34'],
                                     normprof1=pars['norm'],
                                     normprof2=pars['norm'],
                                     normprof3=pars['norm'],
                                     normprof4=pars['norm'],
                                     p_of_k_a=pars['pk'],
                                     lk_arr=np.log(k_arr), a_arr=a_arr,
                                     )
    tk = tkk.eval(0.1, 0.5)
    assert np.all(np.isfinite(tk))


def test_tkkssc_errors():

    hmc = ccl.halos.HMCalculator(COSMO, HMF, HBF, mass_def=M200)
    k_arr = KK
    a_arr = np.array([0.3, 0.5, 0.7, 1.0])

    # Wrong first profile
    with pytest.raises(TypeError):
        ccl.halos.halomod_Tk3D_SSC(COSMO, hmc, None)
    # Wrong other profiles
    for i in range(2, 4):
        kw = {'prof%d' % i: PKC}
        with pytest.raises(TypeError):
            ccl.halos.halomod_Tk3D_SSC(COSMO, hmc, P1, **kw)
    # Wrong 2pts
    with pytest.raises(TypeError):
        ccl.halos.halomod_Tk3D_SSC(COSMO, hmc, P1,
                                   prof12_2pt=P2)
    with pytest.raises(TypeError):
        ccl.halos.halomod_Tk3D_SSC(COSMO, hmc, P1,
                                   prof34_2pt=P2)

    # No normalization for number counts profile
    with pytest.raises(ValueError):
        ccl.halos.halomod_Tk3D_SSC(COSMO, hmc, P2, normprof1=False)
    with pytest.raises(ValueError):
        ccl.halos.halomod_Tk3D_SSC(COSMO, hmc, P1, prof2=P2, normprof2=False)
    with pytest.raises(ValueError):
        ccl.halos.halomod_Tk3D_SSC(COSMO, hmc, P1, prof3=P2, normprof3=False)
    with pytest.raises(ValueError):
        ccl.halos.halomod_Tk3D_SSC(COSMO, hmc, P1, prof4=P2, normprof4=False)

    # Negative profile in logspace
    assert_warns(ccl.CCLWarning, ccl.halos.halomod_Tk3D_1h,
                 COSMO, hmc, P3, prof2=Pneg,
                 lk_arr=np.log(k_arr), a_arr=a_arr,
                 use_log=True)


@pytest.mark.parametrize('kwargs', [
                         # All is_number_counts = False
                         {'prof1': P1, 'prof2': P1,
                          'prof3': P1, 'prof4': P1},
                         #
                         {'prof1': P2, 'prof2': P1,
                          'prof3': P1, 'prof4': P1},
                         #
                         {'prof1': P2, 'prof2': P2,
                          'prof3': P1, 'prof4': P1},
                         #
                         {'prof1': P2, 'prof2': P2,
                          'prof3': P2, 'prof4': P1},
                         #
                         {'prof1': P1, 'prof2': P1,
                          'prof3': P2, 'prof4': P2},
                         #
                         {'prof1': P2, 'prof2': P1,
                          'prof3': P2, 'prof4': P2},
                         #
                         {'prof1': P2, 'prof2': P2,
                          'prof3': P1, 'prof4': P2},
                         #
                         {'prof1': P2, 'prof2': None,
                          'prof3': P1, 'prof4': None},
                         #
                         {'prof1': P2, 'prof2': None,
                          'prof3': None, 'prof4': None},
                         #
                         {'prof1': P2, 'prof2': P2,
                          'prof3': None, 'prof4': None},
                         # As in benchmarks/test_covariances.py
                         {'prof1': P1, 'prof2': P1,
                          'prof3': None, 'prof4': None},
                         # Setting prof34_2pt
                         {'prof1': P1, 'prof2': P1,
                          'prof3': None, 'prof4': None,
                          'prof34_2pt': PKC},
                         # All is_number_counts = True
                         {'prof1': P2, 'prof2': P2,
                          'prof3': P2, 'prof4': P2},

                         ]
                         )
def test_tkkssc_counterterms_gc(kwargs):
    hmc = ccl.halos.HMCalculator(COSMO, HMF, HBF, mass_def=M200,
                                 nlog10M=2)
    k_arr = KK
    a_arr = np.array([0.3, 0.5, 0.7, 1.0])

    # Tk's without clustering terms. Set is_number_counts=False for HOD
    # profiles
    # Ensure HOD profiles are normalized
    kwargs_nogc = kwargs.copy()
    keys = list(kwargs.keys())
    for k in keys:
        v = kwargs[k]
        if isinstance(v, ccl.halos.HaloProfileHOD):
            kwargs_nogc[k] = P2_nogc
            kwargs_nogc['norm' + k] = True
            kwargs['norm' + k] = True
    tkk_nogc = ccl.halos.halomod_Tk3D_SSC(COSMO, hmc,
                                          lk_arr=np.log(k_arr), a_arr=a_arr,
                                          **kwargs_nogc)
    _, _, _, tkk_nogc_arrs = tkk_nogc.get_spline_arrays()
    tk_nogc_12, tk_nogc_34 = tkk_nogc_arrs

    # Tk's with clustering terms
    tkk_gc = ccl.halos.halomod_Tk3D_SSC(COSMO, hmc,
                                        lk_arr=np.log(k_arr), a_arr=a_arr,
                                        **kwargs)
    _, _, _, tkk_gc_arrs = tkk_gc.get_spline_arrays()
    tk_gc_12, tk_gc_34 = tkk_gc_arrs

    # Update the None's to their corresponding values
    if kwargs['prof2'] is None:
        kwargs['prof2'] = kwargs['prof1']
    if kwargs['prof3'] is None:
        kwargs['prof3'] = kwargs['prof1']
    if kwargs['prof4'] is None:
        kwargs['prof4'] = kwargs['prof3']

    # Tk's of the clustering terms
    tkc12 = []
    tkc34 = []
    for aa in a_arr:
        tkc12.append(get_ssc_counterterm_gc(k_arr, aa, hmc, kwargs['prof1'],
                                            kwargs['prof2'], PKC))
        tkc34.append(get_ssc_counterterm_gc(k_arr, aa, hmc, kwargs['prof3'],
                                            kwargs['prof4'], PKC))
    tkc12 = np.array(tkc12)
    tkc34 = np.array(tkc34)

    assert np.abs((tk_nogc_12 - tkc12) / tk_gc_12 - 1).max() < 1e-5
    assert np.abs((tk_nogc_34 - tkc34) / tk_gc_34 - 1).max() < 1e-5


@pytest.mark.parametrize('kwargs', [{f'is_number_counts{i+1}': nc[i] for i in
                                     range(4)} for nc in
                                    itertools.product([True, False],
                                                      repeat=4)])
def test_tkkssc_linear_bias(kwargs):
    hmc = ccl.halos.HMCalculator(COSMO, HMF, HBF, mass_def=M200,
                                 nlog10M=2)
    k_arr = KK
    a_arr = np.array([0.3, 0.5, 0.7, 1.0])

    # Tk's exact version
    prof = ccl.halos.HaloProfileNFW(ccl.halos.ConcentrationDuffy08(M200),
                                    fourier_analytic=True)
    bias1 = 2
    bias2 = 3
    bias3 = 4
    bias4 = 5
    is_nc = False

    # Tk's from tkkssc_linear
    tkk_lin = ccl.halos.halomod_Tk3D_SSC_linear_bias(COSMO, hmc, prof=prof,
                                                     bias1=bias1,
                                                     bias2=bias2,
                                                     bias3=bias3,
                                                     bias4=bias4,
                                                     is_number_counts1=is_nc,
                                                     is_number_counts2=is_nc,
                                                     is_number_counts3=is_nc,
                                                     is_number_counts4=is_nc,
                                                     lk_arr=np.log(k_arr),
                                                     a_arr=a_arr)
    _, _, _, tkk_lin_arrs = tkk_lin.get_spline_arrays()
    tk_lin_12, tk_lin_34 = tkk_lin_arrs
    # Remove the biases
    tk_lin_12 /= (bias1 * bias2)
    tk_lin_34 /= (bias3 * bias4)
    assert np.abs(tk_lin_12 / tk_lin_34 - 1).max() < 1e-5

    # True Tk's (biases for NFW ~ 1)
    tkk = ccl.halos.halomod_Tk3D_SSC(COSMO, hmc, prof1=prof,
                                     lk_arr=np.log(k_arr), a_arr=a_arr,
                                     normprof1=True, normprof2=True,
                                     normprof3=True, normprof4=True)
    _, _, _, tkk_arrs = tkk.get_spline_arrays()
    tk_12, tk_34 = tkk_arrs

    assert np.abs(tk_lin_12 / tk_12 - 1).max() < 1e-2
    assert np.abs(tk_lin_34 / tk_34 - 1).max() < 1e-2

    # Now with clustering
    tkk_lin_nc = ccl.halos.halomod_Tk3D_SSC_linear_bias(COSMO, hmc, prof=prof,
                                                        bias1=bias1,
                                                        bias2=bias2,
                                                        bias3=bias3,
                                                        bias4=bias4,
                                                        **kwargs,
                                                        lk_arr=np.log(k_arr),
                                                        a_arr=a_arr)
    _, _, _, tkk_lin_nc_arrs = tkk_lin_nc.get_spline_arrays()
    tk_lin_nc_12, tk_lin_nc_34 = tkk_lin_nc_arrs
    tk_lin_nc_12 /= (bias1 * bias2)
    tk_lin_nc_34 /= (bias3 * bias4)

    factor12 = 0
    factor12 += bias1 if kwargs['is_number_counts1'] else 0
    factor12 += bias2 if kwargs['is_number_counts2'] else 0

    factor34 = 0
    factor34 += bias3 if kwargs['is_number_counts3'] else 0
    factor34 += bias4 if kwargs['is_number_counts4'] else 0

    tk_lin_ct_12 = np.zeros_like(tk_lin_nc_12)
    if factor12 != 0:
        tk_lin_ct_12 = (tk_lin_12 - tk_lin_nc_12) / factor12  # = pk+i02

    tk_lin_ct_34 = np.zeros_like(tk_lin_nc_34)
    if factor34 != 0:
        tk_lin_ct_34 = (tk_lin_34 - tk_lin_nc_34) / factor34  # = pk+i02

    if (factor12 != 0) and (factor34 != 0):
        assert np.abs(tk_lin_ct_12 / tk_lin_ct_34 - 1).max() < 1e-5

    # True counter terms
    tkc12 = []
    tkc34 = []
    prof.is_number_counts = True  # Trick the function below
    for aa in a_arr:
        # Divide by 2 to account for ~(1 + 1)
        tkc_ia = get_ssc_counterterm_gc(k_arr, aa, hmc, prof, prof, PKC,
                                        normalize=True) / 2
        if factor12 == 0:
            tkc12.append(np.zeros_like(tkc_ia))
        else:
            tkc12.append(tkc_ia)

        if factor34 == 0:
            tkc34.append(np.zeros_like(tkc_ia))
        else:
            tkc34.append(tkc_ia)

    tkc12 = np.array(tkc12)
    tkc34 = np.array(tkc34)

    # Add 1e-100 for the cases when the counter terms are 0
    assert np.abs((tk_lin_ct_12 + 1e-100) / (tkc12 + 1e-100) - 1).max() < 1e-2
    assert np.abs((tk_lin_ct_34 + 1e-100) / (tkc34 + 1e-100) - 1).max() < 1e-2


def test_tkkssc_linear_bias_smoke_and_errors():
    hmc = ccl.halos.HMCalculator(COSMO, HMF, HBF, mass_def=M200,
                                 nlog10M=2)
    k_arr = KK
    a_arr = np.array([0.3, 0.5, 0.7, 1.0])

    # Tk's exact version
    prof = ccl.halos.HaloProfileNFW(ccl.halos.ConcentrationDuffy08(M200),
                                    fourier_analytic=True)

    ccl.halos.halomod_Tk3D_SSC_linear_bias(COSMO, hmc, prof=prof,
                                           p_of_k_a='linear')

    ccl.halos.halomod_Tk3D_SSC_linear_bias(COSMO, hmc, prof=prof,
                                           p_of_k_a='nonlinear')

    pk = COSMO.get_nonlin_power()
    ccl.halos.halomod_Tk3D_SSC_linear_bias(COSMO, hmc, prof=prof, p_of_k_a=pk)

    # Error when prof is not NFW
    with pytest.raises(TypeError):
        ccl.halos.halomod_Tk3D_SSC_linear_bias(COSMO, hmc, prof=P2)

    # Error when p_of_k_a is wrong
    with pytest.raises(TypeError):
        ccl.halos.halomod_Tk3D_SSC_linear_bias(COSMO, hmc, prof=prof,
                                               p_of_k_a=P1)

    # Negative profile in logspace
    assert_warns(ccl.CCLWarning, ccl.halos.halomod_Tk3D_SSC_linear_bias,
                 COSMO, hmc, prof, bias1=-1,
                 lk_arr=np.log(k_arr), a_arr=a_arr,
                 use_log=True)



import numpy as np
import pytest
import pyccl as ccl


COSMO = ccl.Cosmology(
    Omega_c=0.27, Omega_b=0.045, h=0.67, sigma8=0.8, n_s=0.96,
    transfer_function='bbks', matter_power_spectrum='linear')
HBFS = [ccl.halos.HaloBiasSheth99,
        ccl.halos.HaloBiasSheth01,
        ccl.halos.HaloBiasTinker10,
        ccl.halos.HaloBiasBhattacharya11]
MS = [1E13, [1E12, 1E15], np.array([1E12, 1E15])]
MFOF = ccl.halos.MassDef('fof', 'matter')
MVIR = ccl.halos.MassDef('vir', 'critical')
MDFS = [MVIR, MVIR, MFOF, MVIR]


@pytest.mark.parametrize('bM_class', HBFS)
def test_bM_subclasses_smoke(bM_class):
    bM = bM_class(COSMO)
    for m in MS:
        b = bM.get_halo_bias(COSMO, m, 0.9)
        assert np.all(np.isfinite(b))
        assert np.shape(b) == np.shape(m)


@pytest.mark.parametrize('bM_pair', zip(HBFS, MDFS))
def test_bM_mdef_raises(bM_pair):
    bM_class, mdef = bM_pair
    with pytest.raises(ValueError):
        bM_class(COSMO, mdef)


def test_bM_SO_allgood():
    bM = ccl.halos.HaloBiasTinker10(COSMO, MVIR)
    for m in MS:
        b = bM.get_halo_bias(COSMO, m, 0.9)
        assert np.all(np.isfinite(b))
        assert np.shape(b) == np.shape(m)


@pytest.mark.parametrize('name', ['Tinker10', 'Sheth99'])
def test_bM_from_string(name):
    bM_class = ccl.halos.HaloBias.from_name(name)
    assert bM_class == ccl.halos.halo_bias_from_name(name)
    bM = bM_class(COSMO)
    for m in MS:
        b = bM.get_halo_bias(COSMO, m, 0.9)
        assert np.all(np.isfinite(b))
        assert np.shape(b) == np.shape(m)


def test_bM_from_string_raises():
    with pytest.raises(ValueError):
        ccl.halos.HaloBias.from_name('Tinker11')


def test_bM_default():
    bM = ccl.halos.HaloBias(COSMO)
    with pytest.raises(NotImplementedError):
        bM._get_bsigma(COSMO, 1., 1.)

    M_in = 1E12
    lM_out = bM._get_consistent_mass(COSMO,
                                     M_in, 1., bM.mdef)
    assert np.fabs(np.log10(M_in) - lM_out) < 1E-10



from . import ccllib as lib

from .pyutils import check, _get_spline2d_arrays, _get_spline3d_arrays
import numpy as np


class Tk3D(object):
    """A container for \"isotropized\" connected trispectra relevant for
    covariance matrix calculations. I.e. functions of 3 variables of the
    form :math:`T(k_1,k_2,a)`, where :math:`k_i` are wave vector moduli
    and :math:`a` is the scale factor. This function can be provided as
    a 3D array (one dimension per variable), or as two 2D arrays
    corresponding to functions :math:`f_i(k,a)` such that

    .. math::
        T(k_1,k_2,a) = f_1(k_1,a)\\,f_2(k_2,a)

    Typical uses for these objects will be:

    * To store perturbation theory or halo model \"isotropized\"
      connected trispectra of the form:

      .. math::
          \\bar{T}_{abcd}(k_1, k_2, a) = \\int \\frac{d\\varphi_1}{2\\pi}
          \\int \\frac{d\\varphi_2}{2\\pi}
          T_{abcd}({\\bf k_1},-{\\bf k_1},{\\bf k_2},-{\\bf k_2}),

      where :math:`{\bf k}_i\\equiv k_i(\\cos\\varphi_i,\\sin\\varphi_i,0)`,
      and :math:`T_{abcd}({\\bf k}_a,{\\bf k}_b,{\\bf k}_c,{\\bf k}_d)` is
      the connected trispectrum of fields :math:`\\{a,b,c,d\\}`.

    * To store the kernel for super-sample covariance calculations as a
      product of the responses of power spectra to long-wavelength
      overdensity modes :math:`\\delta_L`:

    .. math::
        \\bar{T}_{abcd}(k_1,k_2,a)=
        \\frac{\\partial P_{ab}(k_1,a)}{\\partial\\delta_L}\\,
        \\frac{\\partial P_{cd}(k_2,a)}{\\partial\\delta_L}.

    These objects can then be used, analogously to
    :class:`~pyccl.pk2d.Pk2D` objects, to construct the non-Gaussian
    covariance of angular power spectra via Limber integration.

    Args:
        a_arr (array): an array holding values of the scale factor. Note
            that the trispectrum will be extrapolated as constant on
            values of the scale factor outside those held by this array.
        lk_arr (array): an array holding values of the natural logarithm
            of the wavenumber (in units of Mpc^-1).
        tkk_arr (array): a 3D array with shape `[na,nk,nk]`, where `na`
            and `nk` are the sizes of `a_arr` and `lk_arr` respectively.
            This array should contain the values of the trispectrum
            at the values of scale factor and wavenumber held by `a_arr`
            and `lk_arr`. The array can hold the values of the natural
            logarithm of the trispectrum, depending on the value of
            `is_logt`. If `tkk_arr` is `None`, then it is assumed that
            the trispectrum can be factorized as described above, and
            the two functions :math:`f_i(k_i,a)` are described by
            `pk1_arr` and `pk2_arr`. You are responsible of making sure
            all these arrays are sufficiently well sampled (i.e. the
            resolution of `a_arr` and `lk_arr` is high enough to sample
            the main features in the trispectrum). For reference, CCL
            will use bicubic interpolation to evaluate the trispectrum
            in the 2D space of wavenumbers :math:`(k_1,k_2)` at a fixed
            scale factor, and will use linear interpolation in the
            scale factor dimension.
        pk1_arr (array): a 2D array with shape `[na,nk]` describing the
            first function :math:`f_1(k,a)` that makes up a factorizable
            trispectrum :math:`T(k_1,k_2,a)=f_1(k_1,a)f_2(k_2,a)`.
            `pk1_arr` and `pk2_arr` are ignored if `tkk_arr` is not
            `None`.
        pk2_arr (array): a 2D array with shape `[na,nk]` describing the
            second factor :math:`f_2(k,a)` for a factorizable trispectrum.
        extrap_order_lok (int): extrapolation order to be used on k-values
            below the minimum of the splines (use 0 or 1). Note that
            the extrapolation will be done in either
            :math:`\\log(T(k_1,k_2,a)` or :math:`T(k_1,k_2,a)`,
            depending on the value of `is_logt`.
        extrap_order_hik (int): same as `extrap_order_lok` for
            k-values above the maximum of the splines.
        is_logt (boolean): if True, `tkk_arr`/`pk1_arr`/`pk2_arr` hold the
            natural logarithm of the trispectrum (or its factors).
            Otherwise, the true values of the corresponding quantities are
            expected. Note that arrays will be interpolated in log space
            if `is_logt` is set to `True`.
    """
    def __init__(self, a_arr, lk_arr, tkk_arr=None,
                 pk1_arr=None, pk2_arr=None, extrap_order_lok=1,
                 extrap_order_hik=1, is_logt=True):
        na = len(a_arr)
        nk = len(lk_arr)

        if not np.all(a_arr[1:]-a_arr[:-1] > 0):
            raise ValueError("`a_arr` must be strictly increasing")

        if not np.all(lk_arr[1:]-lk_arr[:-1] > 0):
            raise ValueError("`lk_arr` must be strictly increasing")

        if ((extrap_order_hik not in (0, 1)) or
                (extrap_order_lok not in (0, 1))):
            raise ValueError("Only constant or linear extrapolation in "
                             "log(k) is possible (`extrap_order_hik` or "
                             "`extrap_order_lok` must be 0 or 1).")
        status = 0

        if tkk_arr is None:
            if pk2_arr is None:
                pk2_arr = pk1_arr
            if (pk1_arr is None) or (pk2_arr is None):
                raise ValueError("If trispectrum is factorizable "
                                 "you must provide the two factors")
            if (pk1_arr.shape != (na, nk)) or (pk2_arr.shape != (na, nk)):
                raise ValueError("Input trispectrum factor "
                                 "shapes are wrong")

            self.tsp, status = lib.tk3d_new_factorizable(lk_arr, a_arr,
                                                         pk1_arr.flatten(),
                                                         pk2_arr.flatten(),
                                                         int(extrap_order_lok),
                                                         int(extrap_order_lok),
                                                         int(is_logt), status)
        else:
            if tkk_arr.shape != (na, nk, nk):
                raise ValueError("Input trispectrum shape is wrong")

            self.tsp, status = lib.tk3d_new_from_arrays(lk_arr, a_arr,
                                                        tkk_arr.flatten(),
                                                        int(extrap_order_lok),
                                                        int(extrap_order_lok),
                                                        int(is_logt), status)
        check(status)

    @property
    def has_tsp(self):
        return 'tsp' in vars(self)

    def eval(self, k, a):
        """Evaluate trispectrum. If `k` is a 1D array with size `nk`, the
        output `out` will be a 2D array with shape `[nk,nk]` holding
        `out[i,j] = T(k[j],k[i],a)`, where `T` is the trispectrum function
        held by this `Tk3D` object.

        Args:
            k (float or array_like): wavenumber value(s) in units of Mpc^-1.
            a (float): value of the scale factor

        Returns:
            float or array_like: value(s) of the trispectrum.
        """
        status = 0

        if np.ndim(a) != 0:
            raise TypeError("a must be a floating point number")

        if isinstance(k, int):
            k = float(k)
        if isinstance(k, float):
            f, status = lib.tk3d_eval_single(self.tsp, np.log(k), a, status)
        else:
            k_use = np.atleast_1d(k)
            nk = k_use.size
            f, status = lib.tk3d_eval_multi(self.tsp, np.log(k_use),
                                            a, nk*nk, status)
            f = f.reshape([nk, nk])
        check(status)
        return f

    def __call__(self, k, a):
        """Callable vectorized instance."""
        out = np.array([self.eval(k, aa)
                        for aa in np.atleast_1d(a).astype(float)])
        return out.squeeze()[()]

    def __del__(self):
        if hasattr(self, 'has_tsp'):
            if self.has_tsp and hasattr(self, 'tsp'):
                lib.f3d_t_free(self.tsp)

    def __bool__(self):
        return self.has_tsp

    def get_spline_arrays(self):
        """Get the spline data arrays.

        Returns:
            a_arr (1D ``numpy.ndarray``):
                Array of scale factors.
            lk_arr1, lk_arr2 (1D ``numpy.ndarray``):
                Arrays of :math:``ln(k)``.
            out (list of ``numpy.ndarray``):
                The trispectrum T(k1, k2, z) or its factors f(k1, z), f(k2, z).
        """
        if not self.has_tsp:
            raise ValueError("Tk3D object does not have data.")

        out = []
        if self.tsp.is_product:
            a_arr, lk_arr1, pk_arr1 = _get_spline2d_arrays(self.tsp.fka_1.fka)
            _, lk_arr2, pk_arr2 = _get_spline2d_arrays(self.tsp.fka_2.fka)
            out.append(pk_arr1)
            out.append(pk_arr2)
        else:
            status = 0
            a_arr, status = lib.get_array(self.tsp.a_arr, self.tsp.na, status)
            check(status)
            lk_arr1, lk_arr2, tkka_arr = _get_spline3d_arrays(self.tsp.tkka,
                                                              self.tsp.na)
            out.append(tkka_arr)

        if self.tsp.is_log:
            # exponentiate in-place
            [np.exp(tk, out=tk) for tk in out]

        return a_arr, lk_arr1, lk_arr2, out



from .profiles import HaloProfile, HaloProfileHOD


class Profile2pt(object):
    """ This class implements the 1-halo 2-point correlator between
    two halo profiles.

    .. math::
        \\langle u_1(k) u_2(k) \\rangle.

    In the simplest case the second-order cumulant is just the product
    of the individual Fourier-space profiles. More complicated cases
    are implemented via the parameters of this class.

    Args:
        r_corr (float):
            Tuning knob for the 1-halo 2-point correlation.
            Scale the correlation by :math:`(1+\\rho_{u_1, u_2})`.
            This is useful when the individual 1-halo terms
            are not fully correlated. Example usecases can be found
            in ``arXiv:1909.09102`` and ``arXiv:2102.07701``.
            Defaults to ``r_corr=0``, returning simply the product
            of the fourier profiles.

    """
    def __init__(self, r_corr=0.):
        self.r_corr = r_corr

    def update_parameters(self, r_corr=None):
        """ Update any of the parameters associated with this 1-halo
        2-point correlator. Any parameter set to `None` won't be updated.
        """
        if r_corr is not None:
            self.r_corr = r_corr

    def fourier_2pt(self, prof, cosmo, k, M, a,
                    prof2=None, mass_def=None):
        """ Return the Fourier-space two-point moment between
        two profiles.

        .. math::
           (1+\\rho_{u_1,u_2})\\langle u_1(k)\\rangle\\langle u_2(k) \\rangle

        Args:
            prof (:class:`~pyccl.halos.profiles.HaloProfile`):
                halo profile for which the second-order moment
                is desired.
            cosmo (:class:`~pyccl.core.Cosmology`):
                a Cosmology object.
            k (float or array_like):
                comoving wavenumber in Mpc^-1.
            M (float or array_like):
                halo mass in units of M_sun.
            a (float):
                scale factor.
            prof2 (:class:`~pyccl.halos.profiles.HaloProfile`):
                second halo profile for which the second-order moment
                is desired. If `None`, the assumption is that you want
                an auto-correlation, and `prof` will be used as `prof2`.
            mass_def (:obj:`~pyccl.halos.massdef.MassDef`):
                a mass definition object.

        Returns:
            float or array_like: second-order Fourier-space
            moment. The shape of the output will be `(N_M, N_k)`
            where `N_k` and `N_m` are the sizes of `k` and `M`
            respectively. If `k` or `M` are scalars, the
            corresponding dimension will be squeezed out on output.
        """
        if not isinstance(prof, HaloProfile):
            raise TypeError("prof must be of type `HaloProfile`")

        uk1 = prof.fourier(cosmo, k, M, a, mass_def=mass_def)

        if prof2 is None:
            uk2 = uk1
        else:
            if not isinstance(prof2, HaloProfile):
                raise TypeError("prof2 must be of type "
                                "`HaloProfile` or `None`")

            uk2 = prof2.fourier(cosmo, k, M, a, mass_def=mass_def)

        return uk1 * uk2 * (1 + self.r_corr)


class Profile2ptHOD(Profile2pt):
    """ This class implements the Fourier-space 1-halo 2-point
    correlator for the HOD profile.

    .. math::
       \\langle n_g^2(k)|M,a\\rangle = \\bar{N}_c(M,a)
       \\left[2f_c(a)\\bar{N}_s(M,a) u_{\\rm sat}(r|M,a)+
       (\\bar{N}_s(M,a) u_{\\rm sat}(r|M,a))^2\\right],

    where all quantities are described in the documentation of
    :class:`~pyccl.halos.profiles.HaloProfileHOD`.
    """
    def fourier_2pt(self, prof, cosmo, k, M, a,
                    prof2=None, mass_def=None):
        """ Returns the Fourier-space two-point moment for the HOD
        profile.

        Args:
            prof (:class:`~pyccl.halos.profiles.HaloProfileHOD`):
                halo profile for which the second-order moment
                is desired.
            cosmo (:class:`~pyccl.core.Cosmology`): a Cosmology object.
            k (float or array_like): comoving wavenumber in Mpc^-1.
            M (float or array_like): halo mass in units of M_sun.
            a (float): scale factor.
            prof2 (:class:`~pyccl.halos.profiles.HaloProfile`):
                second halo profile for which the second-order moment
                is desired. If `None`, the assumption is that you want
                an auto-correlation. Note that only auto-correlations
                are allowed in this case.
            mass_def (:obj:`~pyccl.halos.massdef.MassDef`): a mass
                definition object.

        Returns:
            float or array_like: second-order Fourier-space
            moment. The shape of the output will be `(N_M, N_k)`
            where `N_k` and `N_m` are the sizes of `k` and `M`
            respectively. If `k` or `M` are scalars, the
            corresponding dimension will be squeezed out on output.
        """
        if not isinstance(prof, HaloProfileHOD):
            raise TypeError("prof must be of type `HaloProfileHOD`")

        if prof2 is not None:
            if prof2 is not prof:
                raise ValueError("prof2 must be the same as prof")

        return prof._fourier_variance(cosmo, k, M, a, mass_def)



import numpy as np
from .. import ccllib as lib
from ..core import check
from ..pk2d import Pk2D
from ..power import linear_matter_power, nonlin_matter_power
from ..background import growth_factor
from .tracers import PTTracer


class PTCalculator(object):
    """ This class implements a set of methods that can be
    used to compute the various components needed to estimate
    perturbation theory correlations. These calculations are
    currently based on FAST-PT
    (https://github.com/JoeMcEwen/FAST-PT).

    Args:
        with_NC (bool): set to True if you'll want to use
            this calculator to compute correlations involving
            number counts.
        with_IA(bool): set to True if you'll want to use
            this calculator to compute correlations involving
            intrinsic alignments.
        with_dd(bool): set to True if you'll want to use
            this calculator to compute the one-loop matter power
            spectrum.
        log10k_min (float): decimal logarithm of the minimum
            Fourier scale (in Mpc^-1) for which you want to
            calculate perturbation theory quantities.
        log10k_max (float): decimal logarithm of the maximum
            Fourier scale (in Mpc^-1) for which you want to
            calculate perturbation theory quantities.
        pad_factor (float): fraction of the log(k) interval
            you want to add as padding for FFTLog calculations
            within FAST-PT.
        low_extrap (float): decimal logaritm of the minimum
            Fourier scale (in Mpc^-1) for which FAST-PT will
            extrapolate.
        high_extrap (float): decimal logaritm of the maximum
            Fourier scale (in Mpc^-1) for which FAST-PT will
            extrapolate.
        P_window (array_like or None): 2-element array describing
            the tapering window used by FAST-PT. See FAST-PT
            documentation for more details.
        C_window (float): `C_window` parameter used by FAST-PT
            to smooth the edges and avoid ringing. See FAST-PT
            documentation for more details.
        k_cutoff (float): exponential cutoff scale. All power
            spectra will be multiplied by a cutoff factor of the
            form :math:`\\exp(-(k/k_*)^n)`, where :math:`k_*` is
            the cutoff scale. This may be useful when using the
            resulting power spectra to compute correlation
            functions if some of the PT contributions do not
            fall sufficiently fast on small scales. If `None`
            (default), no cutoff factor will be applied.
        n_exp_cutoff (float): exponent of the cutoff factor (see
            `k_cutoff`).
    """

    def __init__(self, with_NC=False, with_IA=False, with_dd=True,
                 log10k_min=-4, log10k_max=2, nk_per_decade=20,
                 pad_factor=1, low_extrap=-5, high_extrap=3,
                 P_window=None, C_window=.75,
                 k_cutoff=None, n_exp_cutoff=4):
        self.with_dd = with_dd
        self.with_NC = with_NC
        self.with_IA = with_IA
        self.P_window = P_window
        self.C_window = C_window

        to_do = ['one_loop_dd']
        if self.with_NC:
            to_do.append('dd_bias')
        if self.with_IA:
            to_do.append('IA')

        nk_total = int((log10k_max - log10k_min) * nk_per_decade)
        self.ks = np.logspace(log10k_min, log10k_max, nk_total)
        n_pad = int(pad_factor * len(self.ks))
        if k_cutoff is not None:
            self.exp_cutoff = np.exp(-(self.ks/k_cutoff)**n_exp_cutoff)
            self.exp_cutoff = self.exp_cutoff[:, None]
        else:
            self.exp_cutoff = 1

        import fastpt as fpt
        self.pt = fpt.FASTPT(self.ks, to_do=to_do,
                             low_extrap=low_extrap,
                             high_extrap=high_extrap,
                             n_pad=n_pad)
        self.one_loop_dd = None
        self.dd_bias = None
        self.ia_ta = None
        self.ia_tt = None
        self.ia_mix = None

    def update_pk(self, pk):
        """ Update the internal PT arrays.

        Args:
            pk (array_like): linear power spectrum sampled at the
                internal `k` values used by this calculator.
        """
        if pk.shape != self.ks.shape:
            raise ValueError("Input spectrum has wrong shape")
        if self.with_NC:
            self._get_dd_bias(pk)
            self.with_dd = True
        elif self.with_dd:
            self._get_one_loop_dd(pk)
        if self.with_IA:
            self._get_ia_bias(pk)

    def _get_one_loop_dd(self, pk):
        # Precompute quantities needed for one-loop dd
        # power spectra. Only needed if dd_bias is not called.
        self.one_loop_dd = self.pt.one_loop_dd(pk,
                                               P_window=self.P_window,
                                               C_window=self.C_window)

    def _get_dd_bias(self, pk):
        # Precompute quantities needed for number counts
        # power spectra.
        self.dd_bias = self.pt.one_loop_dd_bias_b3nl(pk,
                                                     P_window=self.P_window,
                                                     C_window=self.C_window)
        self.one_loop_dd = self.dd_bias[0:1]

    def _get_ia_bias(self, pk):
        # Precompute quantities needed for intrinsic alignment
        # power spectra.
        self.ia_ta = self.pt.IA_ta(pk,
                                   P_window=self.P_window,
                                   C_window=self.C_window)
        self.ia_tt = self.pt.IA_tt(pk,
                                   P_window=self.P_window,
                                   C_window=self.C_window)
        self.ia_mix = self.pt.IA_mix(pk,
                                     P_window=self.P_window,
                                     C_window=self.C_window)

    def get_pgg(self, Pd1d1, g4,
                b11, b21, bs1, b12, b22, bs2,
                sub_lowk, b3nl1=None, b3nl2=None,
                bk21=None, bk22=None, Pgrad=None):
        """ Get the number counts auto-spectrum at the internal
        set of wavenumbers (given by this object's `ks` attribute)
        and a number of redshift values.

        Args:
            Pd1d1 (array_like): 1-loop matter power spectrum at the
                wavenumber values given by this object's `ks` list.
            g4 (array_like): fourth power of the growth factor at
                a number of redshifts.
            b11 (array_like): 1-st order bias for the first tracer
                being correlated at the same set of input redshifts.
            b21 (array_like): 2-nd order bias for the first tracer
                being correlated at the same set of input redshifts.
            bs1 (array_like): tidal bias for the first tracer
                being correlated at the same set of input redshifts.
            b12 (array_like): 1-st order bias for the second tracer
                being correlated at the same set of input redshifts.
            b22 (array_like): 2-nd order bias for the second tracer
                being correlated at the same set of input redshifts.
            bs2 (array_like): tidal bias for the second tracer
                being correlated at the same set of input redshifts.
            sub_lowk (bool): if True, the small-scale white noise
                contribution will be subtracted.
            b3nl1 (array_like): 3-rd order bias for the first tracer.
                If `None`, this contribution won't be included.
            b3nl2 (array_like): 3-rd order bias for the second tracer.
                If `None`, this contribution won't be included.
            bk21 (array_like): non-local bias for the first tracer.
                If `None`, this contribution won't be included.
            bk22 (array_like): non-local bias for the second tracer.
                If `None`, this contribution won't be included.
            Pgrad (array_like): cross-correlation spectrum between
                the matter overdensity :math:`\\delta` and its
                Laplacian (divided by :math:`-k^2`). Evaluated at
                the wavenumber values given by this object's `ks`
                list. If `None`, `Pd1d1` will be used.

        Returns:
            array_like: 2D array of shape `(N_k, N_z)`, where `N_k` \
                is the size of this object's `ks` attribute, and \
                `N_z` is the size of the input redshift-dependent \
                biases and growth factor.
        """
        Pd1d2 = g4[None, :] * self.dd_bias[2][:, None]
        Pd2d2 = g4[None, :] * self.dd_bias[3][:, None]
        Pd1s2 = g4[None, :] * self.dd_bias[4][:, None]
        Pd2s2 = g4[None, :] * self.dd_bias[5][:, None]
        Ps2s2 = g4[None, :] * self.dd_bias[6][:, None]
        Pd1p3 = g4[None, :] * self.dd_bias[8][:, None]
        if Pgrad is None:
            Pgrad = Pd1d1
        Pd1k2 = Pgrad * (self.ks**2)[:, None]

        if b3nl1 is None:
            b3nl1 = np.zeros_like(g4)
        if b3nl2 is None:
            b3nl2 = np.zeros_like(g4)
        if bk21 is None:
            bk21 = np.zeros_like(g4)
        if bk22 is None:
            bk22 = np.zeros_like(g4)

        s4 = 0.
        if sub_lowk:
            s4 = g4 * self.dd_bias[7]
            s4 = s4[None, :]

        pgg = ((b11*b12)[None, :] * Pd1d1 +
               0.5*(b11*b22 + b12*b21)[None, :] * Pd1d2 +
               0.25*(b21*b22)[None, :] * (Pd2d2 - 2.*s4) +
               0.5*(b11*bs2 + b12*bs1)[None, :] * Pd1s2 +
               0.25*(b21*bs2 + b22*bs1)[None, :] * (Pd2s2 - (4./3.)*s4) +
               0.25*(bs1*bs2)[None, :] * (Ps2s2 - (8./9.)*s4) +
               0.5*(b12*b3nl1+b11*b3nl2)[None, :] * Pd1p3 +
               0.5*(b12*bk21+b11*bk22)[None, :] * Pd1k2)

        return pgg*self.exp_cutoff

    def get_pgi(self, Pd1d1, g4, b1, b2, bs, c1, c2, cd):
        """ Get the number counts - IA cross-spectrum at the
        internal set of wavenumbers (given by this object's
        `ks` attribute) and a number of redshift values.

        .. note:: The full non-linear model for the cross-correlation
                  between number counts and intrinsic alignments is
                  still work in progress in FastPT. As a workaround
                  CCL assumes a non-linear treatment of IAs, but only
                  linearly biased number counts.

        Args:
            Pd1d1 (array_like): 1-loop matter power spectrum at the
                wavenumber values given by this object's `ks` list.
            g4 (array_like): fourth power of the growth factor at
                a number of redshifts.
            b1 (array_like): 1-st order bias for the number counts
                being correlated at the same set of input redshifts.
            b2 (array_like): 2-nd order bias for the number counts
                being correlated at the same set of input redshifts.
            bs (array_like): tidal bias for the number counts
                being correlated at the same set of input redshifts.
            c1 (array_like): 1-st order bias for the IA tracer
                being correlated at the same set of input redshifts.
            c2 (array_like): 2-nd order bias for the IA tracer
                being correlated at the same set of input redshifts.
            cd (array_like): overdensity bias for the IA tracer
                being correlated at the same set of input redshifts.

        Returns:
            array_like: 2D array of shape `(N_k, N_z)`, where `N_k` \
                is the size of this object's `ks` attribute, and \
                `N_z` is the size of the input redshift-dependent \
                biases and growth factor.
        """
        a00e, c00e, a0e0e, a0b0b = self.ia_ta
        a0e2, b0e2, d0ee2, d0bb2 = self.ia_mix

        pgi = b1[None, :] * (c1[None, :] * Pd1d1 +
                             (g4*cd)[None, :] * (a00e + c00e)[:, None] +
                             (g4*c2)[None, :] * (a0e2 + b0e2)[:, None])
        return pgi*self.exp_cutoff

    def get_pgm(self, Pd1d1, g4, b1, b2, bs, b3nl=None,
                bk2=None, Pgrad=None):
        """ Get the number counts - matter cross-spectrum at the
        internal set of wavenumbers (given by this object's `ks`
        attribute) and a number of redshift values.

        Args:
            Pd1d1 (array_like): 1-loop matter power spectrum at the
                wavenumber values given by this object's `ks` list.
            g4 (array_like): fourth power of the growth factor at
                a number of redshifts.
            b1 (array_like): 1-st order bias for the number counts
                tracer being correlated at the same set of input
                redshifts.
            b2 (array_like): 2-nd order bias for the number counts
                tracer being correlated at the same set of input
                redshifts.
            bs (array_like): tidal bias for the number counts
                tracer being correlated at the same set of input
                redshifts.
            b3nl (array_like): 3-rd order bias for the number counts
                tracer being correlated at the same set of input
                redshifts. If `None`, this contribution won't be
                included.
            bk2 (array_like): non-local bias for the number counts
                tracer being correlated at the same set of input
                redshifts. If `None`, this contribution won't be
                included.
            Pgrad (array_like): cross-correlation spectrum between
                the matter overdensity :math:`\\delta` and its
                Laplacian (divided by :math:`-k^2`). Evaluated at
                the wavenumber values given by this object's `ks`
                list. If `None`, `Pd1d1` will be used.

        Returns:
            array_like: 2D array of shape `(N_k, N_z)`, where `N_k` \
                is the size of this object's `ks` attribute, and \
                `N_z` is the size of the input redshift-dependent \
                biases and growth factor.
        """
        Pd1d2 = g4[None, :] * self.dd_bias[2][:, None]
        Pd1s2 = g4[None, :] * self.dd_bias[4][:, None]
        Pd1p3 = g4[None, :] * self.dd_bias[8][:, None]
        if Pgrad is None:
            Pgrad = Pd1d1
        Pd1k2 = Pgrad*(self.ks**2)[:, None]
        if b3nl is None:
            b3nl = np.zeros_like(g4)
        if bk2 is None:
            bk2 = np.zeros_like(g4)

        pgm = (b1[None, :] * Pd1d1 +
               0.5 * b2[None, :] * Pd1d2 +
               0.5 * bs[None, :] * Pd1s2 +
               0.5 * b3nl[None, :] * Pd1p3 +
               0.5 * bk2[None, :] * Pd1k2)

        return pgm*self.exp_cutoff

    def get_pii(self, Pd1d1, g4, c11, c21, cd1,
                c12, c22, cd2, return_bb=False,
                return_both=False):
        """ Get the intrinsic alignment auto-spectrum at the internal
        set of wavenumbers (given by this object's `ks` attribute)
        and a number of redshift values.

        Args:
            Pd1d1 (array_like): 1-loop matter power spectrum at the
                wavenumber values given by this object's `ks` list.
            g4 (array_like): fourth power of the growth factor at
                a number of redshifts.
            c11 (array_like): 1-st order bias for the first tracer
                being correlated at the same set of input redshifts.
            c21 (array_like): 2-nd order bias for the first tracer
                being correlated at the same set of input redshifts.
            cd1 (array_like): overdensity bias for the first tracer
                being correlated at the same set of input redshifts.
            c12 (array_like): 1-st order bias for the second tracer
                being correlated at the same set of input redshifts.
            c22 (array_like): 2-nd order bias for the second tracer
                being correlated at the same set of input redshifts.
            cd2 (array_like): overdensity bias for the second tracer
                being correlated at the same set of input redshifts.
            return_bb (bool): if `True`, the B-mode power spectrum
                will be returned.
            return_both (bool): if `True`, both the E- and B-mode
                power spectra will be returned. Supersedes `return_bb`.

        Returns:
            array_like: 2D array of shape `(N_k, N_z)`, where `N_k` \
                is the size of this object's `ks` attribute, and \
                `N_z` is the size of the input redshift-dependent \
                biases and growth factor.
        """
        a00e, c00e, a0e0e, a0b0b = self.ia_ta
        ae2e2, ab2b2 = self.ia_tt
        a0e2, b0e2, d0ee2, d0bb2 = self.ia_mix

        if return_both:
            return_bb = True

        if return_bb:
            pii_bb = ((cd1*cd2*g4)[None, :] * a0b0b[:, None] +
                      (c21*c22*g4)[None, :] * ab2b2[:, None] +
                      ((cd1*c22 + c21*cd2)*g4)[None, :] * d0bb2[:, None])
            if not return_both:
                pii = pii_bb

        if (not return_bb) or return_both:
            pii = ((c11*c12)[None, :] * Pd1d1 +
                   ((c11*cd2 + c12*cd1)*g4)[None, :] * (a00e + c00e)[:, None] +
                   (cd1*cd2*g4)[None, :] * a0e0e[:, None] +
                   (c21*c22*g4)[None, :] * ae2e2[:, None] +
                   ((c11*c22 + c21*c12)*g4)[None, :] * (a0e2 + b0e2)[:, None] +
                   ((cd1*c22 + cd2*c21)*g4)[None, :] * d0ee2[:, None])

        if return_both:
            return pii*self.exp_cutoff, pii_bb*self.exp_cutoff
        else:
            return pii*self.exp_cutoff

    def get_pim(self, Pd1d1, g4, c1, c2, cd):
        """ Get the intrinsic alignment - matter cross-spectrum at
        the internal set of wavenumbers (given by this object's `ks`
        attribute) and a number of redshift values.

        Args:
            Pd1d1 (array_like): 1-loop matter power spectrum at the
                wavenumber values given by this object's `ks` list.
            g4 (array_like): fourth power of the growth factor at
                a number of redshifts.
            c1 (array_like): 1-st order bias for the IA
                tracer being correlated at the same set of input
                redshifts.
            c2 (array_like): 2-nd order bias for the IA
                tracer being correlated at the same set of input
                redshifts.
            cd (array_like): overdensity bias for the IA
                tracer being correlated at the same set of input
                redshifts.

        Returns:
            array_like: 2D array of shape `(N_k, N_z)`, where `N_k` \
                is the size of this object's `ks` attribute, and \
                `N_z` is the size of the input redshift-dependent \
                biases and growth factor.
        """
        a00e, c00e, a0e0e, a0b0b = self.ia_ta
        a0e2, b0e2, d0ee2, d0bb2 = self.ia_mix

        pim = (c1[None, :] * Pd1d1 +
               (g4*cd)[None, :] * (a00e + c00e)[:, None] +
               (g4*c2)[None, :] * (a0e2 + b0e2)[:, None])
        return pim*self.exp_cutoff

    def get_pmm(self, Pd1d1_lin, g4):
        """ Get the one-loop matter power spectrum.

        Args:
            Pd1d1_lin (array_like): 1-loop linear matter power spectrum
                at the wavenumber values given by this object's
                `ks` list.
            g4 (array_like): fourth power of the growth factor at
                a number of redshifts.

        Returns:
            array_like: 2D array of shape `(N_k, N_z)`, where `N_k` \
                is the size of this object's `ks` attribute, and \
                `N_z` is the size of the input redshift-dependent \
                biases and growth factor.
        """
        P1loop = g4[None, :] * self.one_loop_dd[0][:, None]
        pmm = (Pd1d1_lin + P1loop)
        return pmm*self.exp_cutoff


def get_pt_pk2d(cosmo, tracer1, tracer2=None, ptc=None,
                sub_lowk=False, nonlin_pk_type='nonlinear',
                nonloc_pk_type='nonlinear',
                a_arr=None, extrap_order_lok=1, extrap_order_hik=2,
                return_ia_bb=False, return_ia_ee_and_bb=False,
                return_ptc=False, update_ptc=True):
    """Returns a :class:`~pyccl.pk2d.Pk2D` object containing
    the PT power spectrum for two quantities defined by
    two :class:`~pyccl.nl_pt.tracers.PTTracer` objects.

    .. note:: The full non-linear model for the cross-correlation
              between number counts and intrinsic alignments is
              still work in progress in FastPT. As a workaround
              CCL assumes a non-linear treatment of IAs, but only
              linearly biased number counts.

    Args:
        cosmo (:class:`~pyccl.core.Cosmology`): a Cosmology object.
        tracer1 (:class:`~pyccl.nl_pt.tracers.PTTracer`): the first
            tracer being correlated.
        ptc (:class:`PTCalculator`): a perturbation theory
            calculator.
        tracer2 (:class:`~pyccl.nl_pt.tracers.PTTracer`): the second
            tracer being correlated. If `None`, the auto-correlation
            of the first tracer will be returned.
        sub_lowk (bool): if True, the small-scale white noise
            contribution will be subtracted for number counts
            auto-correlations.
        nonlin_pk_type (str): type of 1-loop matter power spectrum
            to use. 'linear' for linear P(k), 'nonlinear' for the internal
            non-linear power spectrum, 'spt' for standard perturbation
            theory power spectrum. Default: 'nonlinear'.
        nonloc_pk_type (str): type of "non-local" matter power spectrum
            to use (i.e. the cross-spectrum between the overdensity and
            its Laplacian divided by :math:`k^2`). Same options as
            `nonlin_pk_type`. Default: 'nonlinear'.
        a_arr (array): an array holding values of the scale factor
            at which the power spectrum should be calculated for
            interpolation. If `None`, the internal values used by
            `cosmo` will be used.
        extrap_order_lok (int): extrapolation order to be used on
            k-values below the minimum of the splines. See
            :class:`~pyccl.pk2d.Pk2D`.
        extrap_order_hik (int): extrapolation order to be used on
            k-values above the maximum of the splines. See
            :class:`~pyccl.pk2d.Pk2D`.
        return_ia_bb (bool): if `True`, the B-mode power spectrum
            for intrinsic alignments will be returned (if both
            input tracers are of type
            :class:`~pyccl.nl_pt.tracers.PTIntrinsicAlignmentTracer`)
            If `False` (default) E-mode power spectrum is returned.
        return_ia_ee_and_bb (bool): if `True`, the E-mode power spectrum
            for intrinsic alignments will be returned in addition to
            the B-mode one (if both input tracers are of type
            :class:`~pyccl.nl_pt.tracers.PTIntrinsicAlignmentTracer`)
            If `False` (default) E-mode power spectrum is returned.
            Supersedes `return_ia_bb`.
        return_ptc (bool): if `True`, the fastpt object used as the PT
            calculator (ptc) will also be returned. This feature may
            be useful if an input ptc is not specified and one is
            initialized when this function is called. If `False` (default)
            the ptc is not output, whether or not it is initialized as
            part of the function call.
        update_ptc (bool): if `False`, do not recompute the tracer-independent
            perturbative quantitities in the fastpt object.

    Returns:
        :class:`~pyccl.pk2d.Pk2D`: PT power spectrum.
        :class:`~pyccl.nl_pt.power.PTCalculator`: PT Calc [optional]
    """
    if a_arr is None:
        status = 0
        na = lib.get_pk_spline_na(cosmo.cosmo)
        a_arr, status = lib.get_pk_spline_a(cosmo.cosmo, na, status)
        check(status, cosmo=cosmo)

    if tracer2 is None:
        tracer2 = tracer1
    if not isinstance(tracer1, PTTracer):
        raise TypeError("tracer1 must be of type `PTTracer`")
    if not isinstance(tracer2, PTTracer):
        raise TypeError("tracer2 must be of type `PTTracer`")

    if ptc is None:
        with_NC = ((tracer1.type == 'NC')
                   or (tracer2.type == 'NC'))
        with_IA = ((tracer1.type == 'IA')
                   or (tracer2.type == 'IA'))
        with_dd = nonlin_pk_type == 'spt'
        ptc = PTCalculator(with_dd=with_dd,
                           with_NC=with_NC,
                           with_IA=with_IA)
        update_ptc = True
    if not isinstance(ptc, PTCalculator):
        raise TypeError("ptc should be of type `PTCalculator`")

    if (tracer1.type == 'NC') or (tracer2.type == 'NC'):
        if not ptc.with_NC:
            raise ValueError("Need number counts bias, "
                             "but calculator didn't compute it")
    if (tracer1.type == 'IA') or (tracer2.type == 'IA'):
        if not ptc.with_IA:
            raise ValueError("Need intrinsic alignment bias, "
                             "but calculator didn't compute it")
    if nonlin_pk_type == 'spt':
        if not ptc.with_dd:
            raise ValueError("Need 1-loop matter power spectrum, "
                             "but calculator didn't compute it")

    if return_ia_ee_and_bb:
        return_ia_bb = True

    # z
    z_arr = 1. / a_arr - 1
    # P_lin(k) at z=0
    pk_lin_z0 = linear_matter_power(cosmo, ptc.ks, 1.)

    # Linear growth factor
    ga = growth_factor(cosmo, a_arr)
    ga4 = ga**4

    if update_ptc:
        # update the PTC to have the require Pk components
        ptc.update_pk(pk_lin_z0)

    if nonlin_pk_type == 'nonlinear':
        Pd1d1 = np.array([nonlin_matter_power(cosmo, ptc.ks, a)
                          for a in a_arr]).T
    elif nonlin_pk_type == 'linear':
        Pd1d1 = np.array([linear_matter_power(cosmo, ptc.ks, a)
                          for a in a_arr]).T
    elif nonlin_pk_type == 'spt':
        pklin = np.array([linear_matter_power(cosmo, ptc.ks, a)
                          for a in a_arr]).T
        Pd1d1 = ptc.get_pmm(pklin, ga4)
    else:
        raise NotImplementedError("Nonlinear option %s not implemented yet" %
                                  (nonlin_pk_type))

    # Compute non-local power spectrum only if needed
    Pgrad = None
    if (((tracer1.type == 'NC') or (tracer2.type == 'NC')) and
            (nonloc_pk_type != nonlin_pk_type)):
        if nonloc_pk_type == 'nonlinear':
            Pgrad = np.array([nonlin_matter_power(cosmo, ptc.ks, a)
                              for a in a_arr]).T
        elif nonloc_pk_type == 'linear':
            Pgrad = np.array([linear_matter_power(cosmo, ptc.ks, a)
                              for a in a_arr]).T
        elif nonloc_pk_type == 'spt':
            pklin = np.array([linear_matter_power(cosmo, ptc.ks, a)
                              for a in a_arr]).T
            Pgrad = ptc.get_pmm(pklin, ga4)
        else:
            raise NotImplementedError("Non-local option %s "
                                      "not implemented yet" %
                                      (nonloc_pk_type))

    if (tracer1.type == 'NC'):
        b11 = tracer1.b1(z_arr)
        b21 = tracer1.b2(z_arr)
        bs1 = tracer1.bs(z_arr)
        b31 = tracer1.b3nl(z_arr)
        bk21 = tracer1.bk2(z_arr)
        if (tracer2.type == 'NC'):
            b12 = tracer2.b1(z_arr)
            b22 = tracer2.b2(z_arr)
            bs2 = tracer2.bs(z_arr)
            b32 = tracer2.b3nl(z_arr)
            bk22 = tracer2.bk2(z_arr)

            p_pt = ptc.get_pgg(Pd1d1, ga4,
                               b11, b21, bs1, b12, b22, bs2,
                               sub_lowk, b3nl1=b31, b3nl2=b32,
                               bk21=bk21, bk22=bk22,
                               Pgrad=Pgrad)
        elif (tracer2.type == 'IA'):
            c12 = tracer2.c1(z_arr)
            c22 = tracer2.c2(z_arr)
            cd2 = tracer2.cdelta(z_arr)
            p_pt = ptc.get_pgi(Pd1d1, ga4,
                               b11, b21, bs1, c12, c22, cd2)
        elif (tracer2.type == 'M'):
            p_pt = ptc.get_pgm(Pd1d1, ga4,
                               b11, b21, bs1, b3nl=b31, bk2=bk21,
                               Pgrad=Pgrad)
        else:
            raise NotImplementedError("Combination %s-%s not implemented yet" %
                                      (tracer1.type, tracer2.type))
    elif (tracer1.type == 'IA'):
        c11 = tracer1.c1(z_arr)
        c21 = tracer1.c2(z_arr)
        cd1 = tracer1.cdelta(z_arr)
        if (tracer2.type == 'IA'):
            c12 = tracer2.c1(z_arr)
            c22 = tracer2.c2(z_arr)
            cd2 = tracer2.cdelta(z_arr)
            p_pt = ptc.get_pii(Pd1d1, ga4,
                               c11, c21, cd1, c12, c22, cd2,
                               return_bb=return_ia_bb,
                               return_both=return_ia_ee_and_bb)
        elif (tracer2.type == 'NC'):
            b12 = tracer2.b1(z_arr)
            b22 = tracer2.b2(z_arr)
            bs2 = tracer2.bs(z_arr)
            p_pt = ptc.get_pgi(Pd1d1, ga4,
                               b12, b22, bs2, c11, c21, cd1)
        elif (tracer2.type == 'M'):
            p_pt = ptc.get_pim(Pd1d1, ga4,
                               c11, c21, cd1)
        else:
            raise NotImplementedError("Combination %s-%s not implemented yet" %
                                      (tracer1.type, tracer2.type))
    elif (tracer1.type == 'M'):
        if (tracer2.type == 'NC'):
            b12 = tracer2.b1(z_arr)
            b22 = tracer2.b2(z_arr)
            bs2 = tracer2.bs(z_arr)
            b32 = tracer2.b3nl(z_arr)
            bk22 = tracer2.bk2(z_arr)
            p_pt = ptc.get_pgm(Pd1d1, ga4,
                               b12, b22, bs2, b3nl=b32, bk2=bk22,
                               Pgrad=Pgrad)
        elif (tracer2.type == 'IA'):
            c12 = tracer2.c1(z_arr)
            c22 = tracer2.c2(z_arr)
            cd2 = tracer2.cdelta(z_arr)
            p_pt = ptc.get_pim(Pd1d1, ga4,
                               c12, c22, cd2)
        elif (tracer2.type == 'M'):
            p_pt = Pd1d1
        else:
            raise NotImplementedError("Combination %s-%s not implemented yet" %
                                      (tracer1.type, tracer2.type))
    else:
        raise NotImplementedError("Combination %s-%s not implemented yet" %
                                  (tracer1.type, tracer2.type))

    # Once you have created the 2-dimensional P(k) array,
    # then generate a Pk2D object as described in pk2d.py.
    if return_ia_ee_and_bb:
        pt_pk_ee = Pk2D(a_arr=a_arr,
                        lk_arr=np.log(ptc.ks),
                        pk_arr=p_pt[0].T,
                        is_logp=False,
                        extrap_order_lok=extrap_order_lok,
                        extrap_order_hik=extrap_order_hik)
        pt_pk_bb = Pk2D(a_arr=a_arr,
                        lk_arr=np.log(ptc.ks),
                        pk_arr=p_pt[1].T,
                        is_logp=False,
                        extrap_order_lok=extrap_order_lok,
                        extrap_order_hik=extrap_order_hik)
        if return_ptc:
            return pt_pk_ee, pt_pk_bb, ptc
        else:
            return pt_pk_ee, pt_pk_bb
    else:
        pt_pk = Pk2D(a_arr=a_arr,
                     lk_arr=np.log(ptc.ks),
                     pk_arr=p_pt.T,
                     is_logp=False,
                     extrap_order_lok=extrap_order_lok,
                     extrap_order_hik=extrap_order_hik)
        if return_ptc:
            return pt_pk, ptc
        else:
            return pt_pk



from numpy.testing import assert_
import pyccl


def test_cclerror_repr():
    """Check that a CCLError can be built from its repr"""
    e = pyccl.CCLError("blah")
    e2 = eval(repr(e))
    assert_(str(e2) == str(e))
    assert_(e2 == e)


def test_cclerror_not_equal():
    """Check that a CCLError can be built from its repr"""
    e = pyccl.CCLError("blah")
    e2 = pyccl.CCLError("blahh")
    assert_(e is not e2)
    assert_(e != e2)
    assert_(hash(e) != hash(e2))


def test_cclwarning_repr():
    """Check that a CCLWarning can be built from its repr"""
    w = pyccl.CCLWarning("blah")
    w2 = eval(repr(w))
    assert_(str(w2) == str(w))
    assert_(w2 == w)

    v = pyccl.CCLDeprecationWarning("blah")
    v2 = eval(repr(v))
    assert str(v2) == str(v)
    assert v2 == v


def test_cclwarning_not_equal():
    """Check that a CCLWarning can be built from its repr"""
    w = pyccl.CCLWarning("blah")
    w2 = pyccl.CCLWarning("blahh")
    assert_(w is not w2)
    assert_(w != w2)
    assert_(hash(w) != hash(w2))

    v = pyccl.CCLDeprecationWarning("blah")
    v2 = pyccl.CCLDeprecationWarning("blahh")
    assert v is not v2
    assert v != v2
    assert hash(v) != hash(v2)


def test_ccl_deprecationwarning_switch():
    import warnings

    # check that warnings are enabled by default
    with warnings.catch_warnings(record=True) as w:
        warnings.warn("test", pyccl.CCLDeprecationWarning)
    assert w[0].category == pyccl.CCLDeprecationWarning

    # switch off CCL (future) deprecation warnings
    pyccl.CCLDeprecationWarning.disable()
    with warnings.catch_warnings(record=True) as w:
        warnings.warn("test", pyccl.CCLDeprecationWarning)
    assert len(w) == 0

    # switch back on
    pyccl.CCLDeprecationWarning.enable()



# Tracers
from .tracers import (
    PTTracer,
    PTMatterTracer,
    PTNumberCountsTracer,
    PTIntrinsicAlignmentTracer,
    translate_IA_norm,
)

# Power spectra
from .power import (
    PTCalculator,
    get_pt_pk2d,
)


__all__ = (
    'PTTracer',
    'PTMatterTracer',
    'PTNumberCountsTracer',
    'PTIntrinsicAlignmentTracer',
    'translate_IA_norm',
    'PTCalculator',
    'get_pt_pk2d',
)



from .. import ccllib as lib
from ..core import check
from ..background import species_types, rho_x, omega_x
import numpy as np


def mass2radius_lagrangian(cosmo, M):
    """ Returns Lagrangian radius for a halo of mass M.
    The lagrangian radius is defined as that enclosing
    the mass of the halo assuming a homogeneous Universe.

    Args:
        cosmo (:class:`~pyccl.core.Cosmology`): A Cosmology object.
        M (float or array_like): halo mass in units of M_sun.

    Returns:
        float or array_like: lagrangian radius in comoving Mpc.
    """
    M_use = np.atleast_1d(M)
    R = (M_use / (4.18879020479 * rho_x(cosmo, 1, 'matter')))**(1./3.)
    if np.ndim(M) == 0:
        R = R[0]
    return R


def convert_concentration(cosmo, c_old, Delta_old, Delta_new):
    """ Computes the concentration parameter for a different mass definition.
    This is done assuming an NFW profile. The output concentration `c_new` is
    found by solving the equation:

    .. math::
        f(c_{\\rm old}) \\Delta_{\\rm old} = f(c_{\\rm new}) \\Delta_{\\rm new}

    where

    .. math::
        f(x) = \\frac{x^3}{\\log(1+x) - x/(1+x)}.

    Args:
        cosmo (:class:`~pyccl.core.Cosmology`): A Cosmology object.
        c_old (float or array_like): concentration to translate from.
        Delta_old (float): Delta parameter associated to the input
            concentration. See description of the MassDef class.
        Delta_new (float): Delta parameter associated to the output
            concentration.

    Returns:
        float or array_like: concentration parameter for the new
        mass definition.
    """
    status = 0
    c_old_use = np.atleast_1d(c_old)
    c_new, status = lib.convert_concentration_vec(cosmo.cosmo,
                                                  Delta_old, c_old_use,
                                                  Delta_new, c_old_use.size,
                                                  status)
    if np.isscalar(c_old):
        c_new = c_new[0]

    check(status, cosmo=cosmo)
    return c_new


class MassDef(object):
    """Halo mass definition. Halo masses are defined in terms of an overdensity
    parameter :math:`\\Delta` and an associated density :math:`X` (either the
    matter density or the critical density):

    .. math::
        M = \\frac{4 \\pi}{3} \\Delta\\,\\rho_X\\, R^3

    where :math:`R` is the halo radius. This object also holds methods to
    translate between :math:`R` and :math:`M`, and to translate masses between
    different definitions if a concentration-mass relation is provided.

    Args:
        Delta (float): overdensity parameter. Pass 'vir' if using virial
            overdensity.
        rho_type (string): either 'critical' or 'matter'.
        c_m_relation (function, optional): concentration-mass relation.
            Provided as a `Concentration` object, or a string corresponding
            to one of the supported concentration-mass relations.
            If `None`, no c(M) relation will be attached to this mass
            definition (and hence one can't translate into other definitions).
    """
    name = 'default'

    def __init__(self, Delta, rho_type, c_m_relation=None):
        # Check it makes sense
        if (Delta != 'fof') and (Delta != 'vir'):
            if isinstance(Delta, str):
                raise ValueError("Unknown Delta type " + Delta)
            if Delta <= 0:
                raise ValueError("Delta must be a positive number")
        self.Delta = Delta
        # Can only be matter or critical
        if rho_type not in ['matter', 'critical']:
            raise ValueError("rho_type must be either \'matter\' "
                             "or \'critical\'")
        self.rho_type = rho_type
        self.species = species_types[rho_type]
        # c(M) relation
        if c_m_relation is None:
            self.concentration = None
        else:
            self._concentration_init(c_m_relation)

    def __eq__(self, other):
        """ Allows you to compare two mass definitions
        """
        return (self.Delta == other.Delta) and \
            (self.rho_type == other.rho_type)

    def _concentration_init(self, c_m_relation):
        from .concentration import Concentration, concentration_from_name
        if isinstance(c_m_relation, Concentration):
            self.concentration = c_m_relation
        elif isinstance(c_m_relation, str):
            # Grab class
            conc_class = concentration_from_name(c_m_relation)
            # instantiate with this mass definition
            self.concentration = conc_class(mdef=self)
        else:
            raise ValueError("c_m_relation must be `None`, "
                             " a string or a `Concentration` object")

    def get_Delta(self, cosmo, a):
        """ Gets overdensity parameter associated to this mass
        definition.

        Args:
            cosmo (:class:`~pyccl.core.Cosmology`): A Cosmology object.
            a (float): scale factor

        Returns:
            float : value of the overdensity parameter.
        """
        if self.Delta == 'vir':
            status = 0
            D, status = lib.Dv_BryanNorman(cosmo.cosmo, a, status)
            return D
        elif self.Delta == 'fof':
            raise ValueError("FoF masses don't have an associated overdensity."
                             "Nor can they be translated into other masses")
        else:
            return self.Delta

    def get_mass(self, cosmo, R, a):
        """ Translates a halo radius into a mass

        .. math::
            M = \\frac{4 \\pi}{3} \\Delta\\,\\rho_X\\, R^3

        Args:
            cosmo (:class:`~pyccl.core.Cosmology`): A Cosmology object.
            R (float or array_like): halo radius in units of Mpc (physical, not
                comoving).
            a (float): scale factor.

        Returns:
            float or array_like: halo mass in units of M_sun.
        """
        R_use = np.atleast_1d(R)
        Delta = self.get_Delta(cosmo, a)
        M = 4.18879020479 * rho_x(cosmo, a, self.rho_type) * Delta * R_use**3
        if np.ndim(R) == 0:
            M = M[0]
        return M

    def get_radius(self, cosmo, M, a):
        """ Translates a halo mass into a radius

        Args:
            cosmo (:class:`~pyccl.core.Cosmology`): A Cosmology object.
            M (float or array_like): halo mass in units of M_sun.
            a (float): scale factor.

        Returns:
            float or array_like: halo radius in units of Mpc (physical, not
                comoving).
        """
        M_use = np.atleast_1d(M)
        Delta = self.get_Delta(cosmo, a)
        R = (M_use / (4.18879020479 * Delta *
                      rho_x(cosmo, a, self.rho_type)))**(1./3.)
        if np.ndim(M) == 0:
            R = R[0]
        return R

    def _get_concentration(self, cosmo, M, a):
        """ Returns concentration for this mass definition.

        Args:
            cosmo (:class:`~pyccl.core.Cosmology`): A Cosmology object.
            M (float or array_like): halo mass in units of M_sun.
            a (float): scale factor.

        Returns:
            float or array_like: halo concentration.
        """
        if self.concentration is None:
            raise RuntimeError("This mass definition doesn't have "
                               "an associated c(M) relation")
        else:
            return self.concentration.get_concentration(cosmo, M, a)

    def translate_mass(self, cosmo, M, a, m_def_other):
        """ Translate halo mass in this definition into another definition

        Args:
            cosmo (:class:`~pyccl.core.Cosmology`): A Cosmology object.
            M (float or array_like): halo mass in units of M_sun.
            a (float): scale factor.
            m_def_other (:obj:`MassDef`): another mass definition.

        Returns:
            float or array_like: halo masses in new definition.
        """
        if self == m_def_other:
            return M
        else:
            if self.concentration is None:
                raise RuntimeError("This mass definition doesn't have "
                                   "an associated c(M) relation")
            else:
                om_this = omega_x(cosmo, a, self.rho_type)
                D_this = self.get_Delta(cosmo, a) * om_this
                c_this = self._get_concentration(cosmo, M, a)
                R_this = self.get_radius(cosmo, M, a)
                om_new = omega_x(cosmo, a, m_def_other.rho_type)
                D_new = m_def_other.get_Delta(cosmo, a) * om_new
                c_new = convert_concentration(cosmo, c_this, D_this, D_new)
                R_new = c_new * R_this / c_this
                return m_def_other.get_mass(cosmo, R_new, a)

    @classmethod
    def from_name(cls, name):
        """ Return mass definition subclass from name string.

        Args:
            name (string):
                a mass definition name (e.g. '200m' for Delta=200 matter)

        Returns:
            MassDef subclass corresponding to the input name.
        """
        mass_defs = {m.name: m for m in cls.__subclasses__()}

        if name in mass_defs:
            return mass_defs[name]
        else:
            raise ValueError(f"Mass definition {name} not implemented.")


class MassDef200m(MassDef):
    """`MassDef` class for the mass definition with Delta=200 times the matter
    density.

    Args:
        c_m (string): concentration-mass relation.
    """
    name = '200m'

    def __init__(self, c_m='Duffy08'):
        super(MassDef200m, self).__init__(200,
                                          'matter',
                                          c_m_relation=c_m)


class MassDef200c(MassDef):
    """`MassDef` class for the mass definition with Delta=200 times
    the critical density.

    Args:
        c_m (string): concentration-mass relation.
    """
    name = '200c'

    def __init__(self, c_m='Duffy08'):
        super(MassDef200c, self).__init__(200,
                                          'critical',
                                          c_m_relation=c_m)


class MassDef500c(MassDef):
    """`MassDef` class for the mass definition
    with Delta=500 times the critical density.

    Args:
        c_m (string): concentration-mass relation.
    """
    name = '500c'

    def __init__(self, c_m='Ishiyama21'):
        super(MassDef500c, self).__init__(500,
                                          'critical',
                                          c_m_relation=c_m)


class MassDefVir(MassDef):
    """`MassDef` class for the mass definition with Delta=Delta_vir times the
    critical density.

    Args:
        c_m (string): concentration-mass relation.
    """
    name = 'vir'

    def __init__(self, c_m='Klypin11'):
        super(MassDefVir, self).__init__('vir',
                                         'critical',
                                         c_m_relation=c_m)



import numpy as np
from scipy.interpolate import interp1d
from ..pyutils import _check_array_params
from ..background import growth_factor
from ..parameters import physical_constants


def translate_IA_norm(cosmo, z, a1=1.0, a1delta=None, a2=None,
                      Om_m2_for_c2=False, Om_m_fid=0.3):
    """
    Function to convert from a_ia values to c_ia values,
    using the standard convention of Blazek 2019 or the variant used
    by the Dark Energy Survey analysis.

    Args:
        cosmo (:class:`~pyccl.core.Cosmology`): cosmology object.
        z (float or array_like): z value(s) where amplitude is evaluated
        a1 (float or array_like): IA a1 at input z values. Defaults to 1.0
        a1delta (float or array_like): IA a1delta at input z values.
            Defaults to None.
        a2 (float or array_like): IA a2 at input z values.
            Defaults to None.
        Om_m2_for_c2 (bool): True to use the Blazek 2019 convention of
            Om_m^2 scaling. Defaults to False
        Om_m_fid (float): Value for Blazek 2019 scaling. Defaults to 0.3.

    Returns:
        c1 (float or array_like): IA c1 at input z values
        c1delta (float or array_like): IA c1delta at input z values
        c2 (float or array_like): IA c2 at input z values
    """

    def check_input_array(a, name):
        if a is None:
            return

        if np.ndim(a) > 1:
            raise ValueError(name +
                             " should be a scalar or 1D")

        if np.ndim(a) == 1:
            if len(a) != len(z):
                raise ValueError("Both z and " + name +
                                 " should have the same size")

    if np.ndim(z) > 1:
        raise ValueError("z should be a scalar or 1D")
    check_input_array(a1, 'a1')
    check_input_array(a2, 'a2')
    check_input_array(a1delta, 'a1delta')

    Om_m = cosmo['Omega_m']
    rho_crit = physical_constants.RHO_CRITICAL
    c1 = c1delta = c2 = None
    gz = growth_factor(cosmo, 1./(1+z))

    if a1 is not None:
        c1 = -1*a1*5e-14*rho_crit*Om_m/gz

    if a1delta is not None:
        c1delta = -1*a1delta*5e-14*rho_crit*Om_m/gz

    if a2 is not None:
        if Om_m2_for_c2:  # Blazek2019 convention
            c2 = a2*5*5e-14*rho_crit*Om_m**2/(Om_m_fid*gz**2)
        else:  # DES convention
            c2 = a2*5*5e-14*rho_crit*Om_m/(gz**2)

    return c1, c1delta, c2


class PTTracer(object):
    """PTTracers contain the information necessary to describe the
    perturbative, non-linear inhomogeneities associated with
    different physical quantities.

    In essence their main function is to store a set of redshift-
    dependent functions (e.g. perturbation theory biases) needed
    in a perturbation theory framework to provide N-point
    correlations.
    """
    def __init__(self):
        self.biases = {}
        self.type = None
        pass

    def get_bias(self, bias_name, z):
        """Get the value of one of the bias functions at a given
        redshift.

        Args:
            bias_name (str): name of the bias function to return.
            z (float or array_like): redshift.

        Returns:
            float or array_like: bias value at the input redshifts.
        """
        if bias_name not in self.biases:
            raise KeyError("Bias %s not included in this tracer" % bias_name)
        return self.biases[bias_name](z)

    def _get_bias_function(self, b):
        # If None, assume it's zero
        if b is None:
            b = 0

        # If it's a scalar, then assume it's a constant function
        if np.ndim(b) == 0:
            def _const(z):
                if np.ndim(z) == 0:
                    return b
                else:
                    return b * np.ones_like(z)

            return _const
        else:  # Otherwise interpolate
            z, b = _check_array_params(b)
            return interp1d(z, b, bounds_error=False,
                            fill_value=b[-1])


class PTMatterTracer(PTTracer):
    """:class:`PTTracer` representing matter fluctuations.
    """
    def __init__(self):
        self.biases = {}
        self.type = 'M'


class PTNumberCountsTracer(PTTracer):
    """:class:`PTTracer` representing number count fluctuations.
    This is described by 1st and 2nd-order biases and
    a tidal field bias. These are provided as floating
    point numbers or tuples of (reshift,bias) arrays.
    If a number is provided, a constant bias is assumed.
    If `None`, a bias of zero is assumed.

    Args:
        b1 (float or tuple of arrays): a single number or a
            tuple of arrays (z, b(z)) giving the first-order
            bias.
        b2 (float or tuple of arrays): as above for the
            second-order bias.
        bs (float or tuple of arrays): as above for the
            tidal bias.
        b3nl (float or tuple of arrays): as above for the
            third-order bias.
        bk2 (float or tuple of arrays): as above for the
            non-local bias.
    """
    def __init__(self, b1, b2=None, bs=None, b3nl=None, bk2=None):
        self.biases = {}
        self.type = 'NC'

        # Initialize b1
        self.biases['b1'] = self._get_bias_function(b1)
        # Initialize b2
        self.biases['b2'] = self._get_bias_function(b2)
        # Initialize bs
        self.biases['bs'] = self._get_bias_function(bs)
        # Initialize b3nl
        self.biases['b3nl'] = self._get_bias_function(b3nl)
        # Initialize bk2
        self.biases['bk2'] = self._get_bias_function(bk2)

    @property
    def b1(self):
        """Internal first-order bias function.
        """
        return self.biases['b1']

    @property
    def b2(self):
        """Internal second-order bias function.
        """
        return self.biases['b2']

    @property
    def bs(self):
        """Internal tidal bias function.
        """
        return self.biases['bs']

    @property
    def b3nl(self):
        """Internal third-order bias function.
        """
        return self.biases['b3nl']

    @property
    def bk2(self):
        """Internal non-local bias function.
        """
        return self.biases['bk2']


class PTIntrinsicAlignmentTracer(PTTracer):
    """:class:`PTTracer` representing intrinsic alignments.
    This is described by 1st and 2nd-order alignment biases
    and an overdensity bias. These are provided as floating
    point numbers or tuples of (reshift,bias) arrays.
    If a number is provided, a constant bias is assumed.
    If `None`, a bias of zero is assumed.

    Args:
        c1 (float or tuple of arrays): a single number or a
            tuple of arrays (z, c1(z)) giving the first-order
            alignment bias.
        c2 (float or tuple of arrays): as above for the
            second-order alignment bias.
        cdelta (float or tuple of arrays): as above for the
            overdensity bias.
    """
    def __init__(self, c1, c2=None, cdelta=None):

        self.biases = {}
        self.type = 'IA'

        # Initialize c1
        self.biases['c1'] = self._get_bias_function(c1)
        # Initialize c2
        self.biases['c2'] = self._get_bias_function(c2)
        # Initialize cdelta
        self.biases['cdelta'] = self._get_bias_function(cdelta)

    @property
    def c1(self):
        """Internal first-order bias function.
        """
        return self.biases['c1']

    @property
    def c2(self):
        """Internal second-order bias function.
        """
        return self.biases['c2']

    @property
    def cdelta(self):
        """Internal overdensity bias function.
        """
        return self.biases['cdelta']
